# Databricks notebook source
# MAGIC %md
# MAGIC # SST Data Assessment / EDA: [SCHOOL]
# MAGIC
# MAGIC First step in the process of transforming raw (PDP) data into actionable, data-driven insights for advisors: validate the data, perform exploratory data analysis, and confirm alignment and/or clarify options on key technical decisions that will determine next steps.
# MAGIC
# MAGIC ### References
# MAGIC
# MAGIC - [Data science product components (Confluence doc)](https://datakind.atlassian.net/wiki/spaces/TT/pages/237862913/Data+science+product+components+the+modeling+process)
# MAGIC - PDP raw data dictionaries: [Course dataset](https://help.studentclearinghouse.org/pdp/knowledge-base/course-level-analysis-ready-file-data-dictionary) and [Cohort dataset](https://help.studentclearinghouse.org/pdp/knowledge-base/cohort-level-analysis-ready-file-data-dictionary)
# MAGIC - [Databricks runtimes release notes](https://docs.databricks.com/en/release-notes/runtime/index.html)
# MAGIC - [Databricks developer notebook utilities](https://docs.databricks.com/en/dev-tools/databricks-utils.html)
# MAGIC - [SCHOOL WEBSITE](https://example.com)

# COMMAND ----------

# MAGIC %md
# MAGIC # Setup

# COMMAND ----------

# MAGIC %sh python --version

# COMMAND ----------

# install dependencies, most of which should come through our 1st-party SST package
# %pip install git+https://github.com/datakind/student-success-tool.git@develop

# COMMAND ----------

# MAGIC %restart_python

# COMMAND ----------

import logging
import os
import sys

import matplotlib.pyplot as plt
import missingno as msno
import numpy as np
import pandas as pd
import seaborn as sb
from databricks.connect import DatabricksSession

from student_success_tool.analysis import pdp

# COMMAND ----------

logging.basicConfig(level=logging.INFO)
logging.getLogger("py4j").setLevel(logging.WARNING)  # ignore databricks logger

try:
    spark_session = DatabricksSession.builder.getOrCreate()
except Exception:
    logging.warning("unable to create spark session; are you in a Databricks runtime?")
    pass

# COMMAND ----------

# MAGIC %md
# MAGIC ## `student-success-intervention` hacks

# COMMAND ----------

# MAGIC %load_ext autoreload
# MAGIC %autoreload 2

# COMMAND ----------

# HACK: insert our 1st-party (school-specific) code into PATH
if "../" not in sys.path:
    sys.path.insert(1, "../")

# TODO: specify school's subpackage
from analysis import *  # noqa: F403

# COMMAND ----------

# MAGIC %md
# MAGIC ## unity catalog config

# COMMAND ----------

catalog = "sst_dev"

# configure where data is to be read from / written to
inst_name = "SCHOOL"  # TODO: fill in school's name in Unity Catalog
read_schema = f"{inst_name}_bronze"
write_schema = f"{inst_name}_silver"

path_volume = os.path.join(
    "/Volumes", catalog, read_schema, f"{inst_name}_bronze_file_volume"
)
path_table = f"{catalog}.{read_schema}"
print(f"{path_table=}")
print(f"{path_volume=}")

# COMMAND ----------

# MAGIC %md
# MAGIC # Read and Validate Raw Data

# COMMAND ----------

# MAGIC %md
# MAGIC ## course dataset

# COMMAND ----------

# TODO: fill in school's name; may not be same as in the schemas above
fpath_course = os.path.join(path_volume, "SCHOOL_COURSE_AR_DEID_DTTM.csv")

# COMMAND ----------

# read without any schema validation, so we can look at the data "raw"
df_course_raw = pdp.dataio.read_raw_pdp_course_data_from_file(
    fpath_course, schema=None, dttm_format="%Y%m%d.0"
)
print(f"rows x cols = {df_course_raw.shape}")
df_course_raw.head()

# COMMAND ----------

df_course_raw.dtypes.value_counts()

# COMMAND ----------

# MAGIC %md
# MAGIC Quick checks:
# MAGIC - [ ] data exists where it should
# MAGIC - [ ] rows and columns are (roughly) as expected
# MAGIC - [ ] datetimes are correctly parsed (they're consistently, legibly formatted)

# COMMAND ----------

# try to read data while validating with the "base" PDP schema
df_course = pdp.dataio.read_raw_pdp_course_data_from_file(
    fpath_course, schema=pdp.schemas.RawPDPCourseDataSchema, dttm_format="%Y%m%d.0"
)
df_course

# COMMAND ----------

# MAGIC %md
# MAGIC If the above command works, and `df_course` is indeed a `pd.DataFrame` containing the validated + parsed PDP cohort dataset, then you're all set, and can skip ahead to the next section. If not, and this is instead a json blob of schema errors, then you'll need to iteratively develop school-specific overrides. There are existing examples you can refer to in the `student-success-intervention` repo.
# MAGIC
# MAGIC This will involve some ad-hoc exploratory work, depending on the schema errors. For example:
# MAGIC
# MAGIC ```python
# MAGIC # => any dupes? how to handle them??
# MAGIC >>> course_unique_cols = pdp.schemas.RawPDPCourseDataSchema.Config.unique
# MAGIC >>> df_course_raw.groupby(course_unique_cols).size().value_counts()
# MAGIC # categorical values with non-standard values?
# MAGIC >>> df_course_raw["co_requisite_course"].value_counts(dropna=False)
# MAGIC ```
# MAGIC
# MAGIC Depending on the results, you'll need a child data schema specific to SCHOOL and/or a preprocessing function to handle more significant deviations, such as duplicate rows. The former looks something like this:
# MAGIC
# MAGIC ```python
# MAGIC import pandas as pd
# MAGIC import pandera as pda
# MAGIC import pandera.typing as pt
# MAGIC
# MAGIC class RawSCHOOLCourseDataSchema(pdp.schemas.RawPDPCourseDataSchema):
# MAGIC     co_requisite_course: pt.Series[pd.CategoricalDtype] = pda.Field(
# MAGIC         nullable=True, dtype_kwargs={"categories": ["YES", "NO"]}
# MAGIC     )
# MAGIC ```
# MAGIC
# MAGIC The latter could be... many things. But fundamentally, it's a function that accepts a "raw", non-compliant dataframe and outputs a dataframe that is able to be validated and parsed by the schema. For example:
# MAGIC
# MAGIC ```python
# MAGIC def course_data_preprocess_func(df: pd.DataFrame, *, unique_cols: list[str]) -> pd.DataFrame:
# MAGIC     df_preproc = df.drop_duplicates(subset=unique_columns)
# MAGIC     if len(df_preproc) < len(df):
# MAGIC         logging.warning(
# MAGIC             "%s duplicate rows found and dropped",
# MAGIC             len(df) - len(df_preproc),
# MAGIC         )
# MAGIC     return df_preproc
# MAGIC ```
# MAGIC
# MAGIC So, in the worst-case scenario, your validation call will look something like this:
# MAGIC
# MAGIC ```python
# MAGIC import functools as ft
# MAGIC
# MAGIC df_course = pdp.dataio.read_raw_pdp_course_data_from_file(
# MAGIC     fpath_course,
# MAGIC     schema=RawSCHOOLCourseDataSchema,
# MAGIC     dttm_format="%Y%m%d.0",
# MAGIC     preprocess_func=ft.partial(
# MAGIC         course_data_preprocess_func,
# MAGIC         unique_cols=pdp.schemas.RawPDPCourseDataSchema.Config.unique
# MAGIC     ),
# MAGIC )
# MAGIC ```
# MAGIC
# MAGIC At this point, `df_course` should be a properly validated and parsed data frame, ready for exploratory data analysis.


# COMMAND ----------

# MAGIC %md
# MAGIC ## cohort dataset

# COMMAND ----------


# TODO: fill in school's name; may not be same as in the schemas above
fpath_cohort = os.path.join(path_volume, "SCHOOL_COHORT_AR_DEID_DTTM.csv")

# COMMAND ----------

# read without any schema validation, so we can look at the data "raw"
df_cohort_raw = pdp.dataio.read_raw_pdp_cohort_data_from_file(fpath_cohort, schema=None)
print(f"rows x cols = {df_cohort_raw.shape}")
df_cohort_raw.head()

# COMMAND ----------

# try to read data while validating with the "base" PDP schema
df_cohort = pdp.dataio.read_raw_pdp_cohort_data_from_file(
    fpath_cohort, schema=pdp.schemas.base.RawPDPCohortDataSchema
)
df_cohort

# COMMAND ----------

# MAGIC %md
# MAGIC Repeat the process from before, as needed: iteratively develop a school-specific cohort data schema and/or preprocessing function, then use them to produce the the validated/parsed data frame for analysis.

# COMMAND ----------

# MAGIC %md
# MAGIC ## takeaways / questions
# MAGIC
# MAGIC - ...

# COMMAND ----------

# MAGIC %md
# MAGIC ## save validated data

# COMMAND ----------

pdp.dataio.write_data_to_delta_table(
    df_course,
    f"{catalog}.{write_schema}.course_dataset_validated",
    spark_session=spark_session,
)

# COMMAND ----------

pdp.dataio.write_data_to_delta_table(
    df_cohort,
    f"{catalog}.{write_schema}.cohort_dataset_validated",
    spark_session=spark_session,
)

# COMMAND ----------

# MAGIC %md
# MAGIC # Exploratory Data Analysis

# COMMAND ----------

# MAGIC %md
# MAGIC %md
# MAGIC ## read validated data
# MAGIC
# MAGIC (so you don't have to execute the validation process more than once)

# COMMAND ----------

# use base or school-specific schema, as needed
df_course = pdp.schemas.RawPDPCourseDataSchema(
    pdp.dataio.read_data_from_delta_table(
        f"{catalog}.{write_schema}.course_dataset_validated",
        spark_session=spark_session,
    )
)
df_course.shape

# COMMAND ----------

df_cohort = pdp.schemas.RawCohortDataSchema(
    pdp.dataio.read_data_from_delta_table(
        f"{catalog}.{write_schema}.cohort_dataset_validated",
        spark_session=spark_session,
    )
)
df_cohort.shape

# COMMAND ----------

# MAGIC %md
# MAGIC ## summary stats

# COMMAND ----------

# decent, general-purpose summarization of a data frame
dbutils.data.summarize(df_course, precise=False)  # noqa: F405

# COMMAND ----------

# specific follow-ups, for example
# df_course["grade"].value_counts(normalize=True, dropna=False)
# df_course["delivery_method"].value_counts(normalize=True, dropna=False)

# COMMAND ----------

dbutils.data.summarize(df_cohort, precise=True)  # noqa: F405

# COMMAND ----------

# specific follow-ups, for example
# df_course["cohort"].value_counts(normalize=True, dropna=False)
# df_course["enrollment_type"].value_counts(normalize=True, dropna=False)

# COMMAND ----------

# MAGIC %md
# MAGIC ### takeaways / questions
# MAGIC
# MAGIC - ...

# COMMAND ----------

# MAGIC %md
# MAGIC ## data validation

# COMMAND ----------

# MAGIC %md
# MAGIC ### null values

# COMMAND ----------

_ = msno.matrix(
    df_course.sort_values(
        by=["academic_year", "enrolled_at_other_institution_s"], ignore_index=True
    ),
    sparkline=False,
    labels=True,
)

# COMMAND ----------

df_course.isna().mean(axis="index").sort_values(ascending=False)

# COMMAND ----------

# confirm that *any* null course identifier means *all* null course identifiers
# this is a known issue in PDP data -- course data is missing for students that transfer out
# we typically drop these records (see "filter invalid rows" section, below) since there is no info
# and they're not the population of interest
# definitely confirm with SCHOOL stakeholders during data assessment presentation
_ = msno.matrix(
    df_course.loc[df_course["course_prefix"].isna(), :].sort_values(
        by=["academic_year", "student_guid"], ignore_index=True
    ),
    sparkline=False,
    labels=True,
)

# COMMAND ----------

num_missing_course_id = df_course["course_prefix"].isna().sum()
pct_missing_course_id = 100 * num_missing_course_id / len(df_course)
print(
    f"num rows missing course id: {num_missing_course_id} ({pct_missing_course_id:.1f}%)"
)

# COMMAND ----------

# check if null course identifiers are (almost) entirely for enrollments at other insts
df_course.loc[
    df_course["course_prefix"].isna(), "enrolled_at_other_institution_s"
].value_counts()

# COMMAND ----------

# here's the null visualization for rows with students not enrolled at other institutions
_ = msno.matrix(
    (
        df_course[df_course["enrolled_at_other_institution_s"].eq("N")].sort_values(
            by=["academic_year", "student_guid"]
        )
    )
)

# COMMAND ----------

_ = msno.matrix(
    df_cohort.sort_values(by=["cohort", "cohort_term"], ignore_index=True),
    sparkline=False,
    labels=True,
)

# COMMAND ----------

with pd.option_context("display.max_rows", 100):
    print(df_cohort.isna().mean(axis="index").sort_values(ascending=False))

# COMMAND ----------

# MAGIC %md
# MAGIC #### takeaways / questions
# MAGIC
# MAGIC - ...

# COMMAND ----------

# MAGIC %md
# MAGIC ### datasets join

# COMMAND ----------

df = (
    pd.merge(
        df_cohort,
        df_course,
        on="student_guid",
        how="outer",
        suffixes=("_cohort", "_course"),
        indicator=True,
    )
    # HACK: columns overlap on more than just student_guid
    # let's rename/drop a relevant few for convenience
    .rename(
        columns={
            "cohort_cohort": "cohort",
            "cohort_term_cohort": "cohort_term",
            "student_age_cohort": "student_age",
            "race_cohort": "race",
            "ethnicity_cohort": "ethnicity",
            "gender_cohort": "gender",
            "institution_id_cohort": "institution_id",
        }
    )
    .drop(
        columns=[
            "cohort_course",
            "cohort_term_course",
            "student_age_course",
            "race_course",
            "ethnicity_course",
            "gender_course",
            "institution_id_course",
        ]
    )
)
df["_merge"].value_counts()

# COMMAND ----------

# any patterns in mis-joined records?
df.loc[df["_merge"] != "both", :]

# COMMAND ----------

# which students don't appear in both datasets?
df.loc[df["_merge"] != "both", "student_guid"].unique().tolist()

# COMMAND ----------

# MAGIC %md
# MAGIC #### takeaways / questions
# MAGIC
# MAGIC - ...

# COMMAND ----------

# MAGIC %md
# MAGIC ### logical consistency

# COMMAND ----------

# check for "pre-cohort" courses
(
    sb.histplot(
        (
            df.assign(
                has_pre_cohort_courses=lambda df: df["cohort"].gt(df["academic_year"])
            )
            .groupby(by=["student_guid", "cohort"])
            .agg(has_pre_cohort_courses=("has_pre_cohort_courses", "any"))
            .reset_index(drop=False)
            .sort_values("cohort")
        ),
        y="cohort",
        hue="has_pre_cohort_courses",
        multiple="stack",
        shrink=0.75,
        edgecolor="white",
    ).set(xlabel="Number of Students")
)

# COMMAND ----------

df_pre_cohort = df.loc[df["cohort"].gt(df["academic_year"]), :].assign(
    cohort_id=lambda df: df["cohort"].str.cat(df["cohort_term"], sep=" "),
    term_id=lambda df: df["academic_year"].str.cat(df["academic_term"], sep=" "),
)
df_pre_cohort[["student_guid", "cohort_id", "term_id", "enrollment_type"]]

# COMMAND ----------

# MAGIC %md
# MAGIC ### filter invalid rows(?)

# COMMAND ----------

# this is probably a filter you'll want to apply
# these courses known to be an issue w/ PDP data
df_course_valid = df_course.loc[df_course["course_number"].notna(), :]
df_course_valid

# COMMAND ----------

# MAGIC %md
# MAGIC ## plots and stats

# COMMAND ----------

# MAGIC %md
# MAGIC **Note:** You'll probably want to use the "valid" dataframes for most of these plots, but not necessarily for all. For simplicity, all these example plots will just use the base data w/o extra data validation filtering applied. It's your call!

# COMMAND ----------

ax = sb.histplot(
    df_cohort.sort_values("cohort"),
    y="cohort",
    hue="cohort_term",
    multiple="stack",
    shrink=0.75,
    edgecolor="white",
)
_ = ax.set(xlabel="Number of Students")

# COMMAND ----------

num_cohorts = df_cohort["cohort"].nunique()
first_cohort, last_cohort = df_cohort["cohort"].min(), df_cohort["cohort"].max()
print(f"{num_cohorts} cohorts ({first_cohort} through {last_cohort})")

# COMMAND ----------

print(df_cohort["cohort_term"].value_counts(normalize=True, dropna=False), end="\n\n")
print(
    df_cohort["enrollment_type"].value_counts(normalize=True, dropna=False), end="\n\n"
)
print(
    df_cohort["enrollment_intensity_first_term"].value_counts(
        normalize=True, dropna=False
    ),
    end="\n\n",
)
print(
    df_cohort["credential_type_sought_year_1"].value_counts(
        normalize=True, dropna=False
    ),
    end="\n\n",
)

# COMMAND ----------

df_cohort["gpa_group_year_1"].describe()

# COMMAND ----------

ax = sb.histplot(
    df_course.sort_values(by="academic_year"),
    y="academic_year",
    hue="academic_term",
    multiple="stack",
    shrink=0.75,
    edgecolor="white",
)
_ = ax.set(xlabel="Number of Course Enrollments")

# COMMAND ----------

num_ayears = df_course["academic_year"].nunique()
first_ayear, last_ayear = (
    df_course["academic_year"].min(),
    df_course["academic_year"].max(),
)
print(f"{num_ayears} academic years ({first_ayear} through {last_ayear})")

# COMMAND ----------

num_courses = (
    df_course["course_prefix"].str.cat(df_course["course_number"], sep=" ").nunique()
)
num_subjects = df_course["course_cip"].nunique()
print(f"{num_courses} distinct courses, {num_subjects} distinct subjects")

# COMMAND ----------

ax = sb.histplot(
    df_cohort,
    y="enrollment_type",
    hue="enrollment_intensity_first_term",
    multiple="stack",
    discrete=True,
    shrink=0.75,
    edgecolor="white",
)
_ = ax.set(xlabel="Number of Students")

# COMMAND ----------

# same as plot above, only in cross-tab form
100 * pdp.eda.compute_crosstabs(
    df_cohort,
    "enrollment_type",
    "enrollment_intensity_first_term",
    normalize=True,
)

# COMMAND ----------

(
    sb.histplot(
        df_cohort,
        y="gender",
        hue="student_age",
        multiple="stack",
        shrink=0.75,
        edgecolor="white",
    ).set(xlabel="Number of Students")
)

# COMMAND ----------

# MAGIC %md
# MAGIC And so on, and so forth.

# COMMAND ----------

ax = sb.histplot(
    pd.merge(
        df_course.groupby("student_guid")
        .size()
        .rename("num_courses_enrolled")
        .reset_index(drop=False),
        df_cohort[["student_guid", "enrollment_intensity_first_term"]],
        on="student_guid",
        how="inner",
    ),
    x="num_courses_enrolled",
    hue="enrollment_intensity_first_term",
    multiple="stack",
    binwidth=5,
    edgecolor="white",
)
ax.set(xlabel="Number of courses enrolled (total)", ylabel="Number of students")
sb.move_legend(ax, loc="upper right", bbox_to_anchor=(1, 1))

# COMMAND ----------

jg = sb.jointplot(
    df_course.groupby("student_guid").agg(
        {"number_of_credits_attempted": "sum", "number_of_credits_earned": "sum"}
    ),
    x="number_of_credits_attempted",
    y="number_of_credits_earned",
    kind="hex",
    joint_kws={"bins": "log"},
    marginal_kws={"edgecolor": "white"},
    ratio=4,
)
jg.refline(y=120.0)  # or whichever num credits earned is a relavent benchmark
jg.set_axis_labels("Number of Credits Attempted", "Number of Credits Earned")

# COMMAND ----------

# MAGIC %md
# MAGIC And so on, and so forth.

# COMMAND ----------

# MAGIC %md
# MAGIC ### takeaways / questions
# MAGIC
# MAGIC - ...

# COMMAND ----------

# MAGIC %md
# MAGIC ## variable associations

# COMMAND ----------

df_assoc_course = pdp.eda.compute_pairwise_associations(
    df_course,
    exclude_cols=[
        "student_guid",
        "institution_id",
        "student_age",
        "gender",
        "race",
        "ethnicity",
    ],
)
df_assoc_course

# COMMAND ----------

fig, ax = plt.subplots(figsize=(10, 10))
sb.heatmap(
    df_assoc_course.astype(np.float32),
    xticklabels=df_assoc_course.columns,
    yticklabels=df_assoc_course.columns,
    vmin=0.0,
    vmax=1.0,
    ax=ax,
)
_ = ax.set_xticklabels(
    ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor"
)

# COMMAND ----------

df_assoc_cohort = pdp.eda.compute_pairwise_associations(
    df_cohort,
    exclude_cols=[
        "student_guid",
        "institution_id",
        "student_age",
        "gender",
        "race",
        "ethnicity",
    ],
)
df_assoc_course

# COMMAND ----------

fig, ax = plt.subplots(figsize=(10, 10))
sb.heatmap(
    df_assoc_cohort.astype(np.float32),
    xticklabels=df_assoc_cohort.columns,
    yticklabels=df_assoc_cohort.columns,
    vmin=0.0,
    vmax=1.0,
    ax=ax,
)
_ = ax.set_xticklabels(
    ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor"
)

# COMMAND ----------

# MAGIC %md
# MAGIC ### takeaways / questions
# MAGIC
# MAGIC - ...

# COMMAND ----------

# MAGIC %md
# MAGIC # Wrap-up

# COMMAND ----------

# MAGIC %md
# MAGIC - [ ] Add school-specific data schemas and/or preprocessing functions into the appropriate directory in the [`student-success-intervention` repository](https://github.com/datakind/student-success-intervention)
# MAGIC - ...

# COMMAND ----------



import os
import shutil
import typing as t
import uuid
from collections.abc import Sequence

import matplotlib.figure
import matplotlib.pyplot as plt
import mlflow
import mlflow.artifacts
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.calibration import calibration_curve
from sklearn.metrics import recall_score
from statsmodels.nonparametric.smoothers_lowess import lowess

# TODO: eventually we should use the custom_style.mplstyle colors, but currently
# the color palette does not have distinct enough colors for calibration by group
# where there are a lot of subgroups
PALETTE = sns.color_palette("Paired")

PosLabelType = t.Optional[int | float | bool | str]


def extract_training_data_from_model(
    automl_experiment_id: str, data_runname: str = "Training Data Storage and Analysis"
) -> pd.DataFrame:
    """
    Read training data from a model into a pandas DataFrame. This allows us to run more
    evaluations of the model, ensuring that we are using the same train/test/validation split

    Args:
        automl_experiment_id: Experiment ID of the AutoML experiment
        data_runname: The runName tag designating where there training data is stored

    Returns:
        The data used for training a model, with train/test/validation flags
    """
    run_df = mlflow.search_runs(
        experiment_ids=[automl_experiment_id], output_format="pandas"
    )
    assert isinstance(run_df, pd.DataFrame)  # type guard
    data_run_id = run_df[run_df["tags.mlflow.runName"] == data_runname]["run_id"].item()

    # Create temp directory to download input data from MLflow
    input_temp_dir = os.path.join(
        os.environ["SPARK_LOCAL_DIRS"], "tmp", str(uuid.uuid4())[:8]
    )
    os.makedirs(input_temp_dir)

    # Download the artifact and read it into a pandas DataFrame
    input_data_path = mlflow.artifacts.download_artifacts(
        run_id=data_run_id, artifact_path="data", dst_path=input_temp_dir
    )
    df_loaded = pd.read_parquet(os.path.join(input_data_path, "training_data"))
    # Delete the temp data
    shutil.rmtree(input_temp_dir)

    return df_loaded


def create_risk_score_histogram(
    risk_score: str | Sequence, title_suffix: str
) -> matplotlib.figure.Figure:
    """
    Create histogram of risk scores

    Args:
        risk_score: risk scores
        title_suffix: suffix for plot title
    """
    fig, ax = plt.subplots()
    sns.histplot(x=risk_score, ax=ax, color=PALETTE[1])
    ax.set(xlabel="Risk Score", title=f"Distribution of risk scores - {title_suffix}")
    return fig


def check_array_of_arrays(input_array: pd.Series) -> bool:
    """
    Check if an input array contains sub-arrays. Used for plotting different
    groups of predictions

    Args:
        input_array

    Returns:
        True if the input_array contains sub-arrays
    """
    try:
        assert isinstance(input_array, pd.Series)
        assert isinstance(input_array[0], list)
        return True
    except Exception:
        return False


def create_calibration_curve(
    y_true: pd.Series,
    risk_score: pd.Series,
    keys: str | list[str],
    title_suffix: str,
    pos_label: PosLabelType,
    lowess_frac: t.Optional[float] = None,
) -> matplotlib.figure.Figure:
    """
    Create calibration plot

    Args:
        y_true (array-like of shape (n_samples,) or (n_groups,)): overall or group-level true outcome class
        risk_score (array-like of shape (n_samples,) or (n_groups,)): overall or group level predicted risk scores
        keys: overall or subgroup level labels for labeling lines
        title_suffix: suffix for plot title
        pos_label: label identifying the positive class. Defaults to True.

    Returns:
        line plot of prediction bins X fraction of positive class
    """
    if not check_array_of_arrays(y_true):
        y_true = [y_true]
        risk_score = [risk_score]
        keys = [keys]  # type: ignore

    fig, ax = plt.subplots()

    for j in range(len(y_true)):
        prob_true, prob_pred = calibration_curve(
            y_true[j],
            risk_score[j],
            n_bins=10,
            strategy="uniform",
            pos_label=pos_label,  # type: ignore
        )
        if lowess_frac:
            # When we create calibration curves with less data points (i.e. for subgroups),
            # it can look choppy and be difficult to interpret. We use locally weighted
            # scatterplot smoothing (LOWESS) to fit a smooth curve to the data points. LOWESS
            # is a non-parametric smoother that tries to find a curve of best fit without
            # assuming the data fits a particular shape.
            # Resource: https://www.statisticshowto.com/lowess-smoothing/
            prob_true = lowess(
                endog=prob_true,
                exog=prob_pred,
                frac=lowess_frac,
                it=0,
                is_sorted=False,
                return_sorted=False,
            )
        sns.lineplot(
            x=prob_pred, y=prob_true, color=PALETTE[j + 1], ax=ax, label=keys[j]
        )
    sns.lineplot(
        x=[0, 1],
        y=[0, 1],
        linestyle="dashed",
        color=PALETTE[0],
        ax=ax,
        label="Perfectly calibrated",
    )

    ax.set(
        xlabel="Mean predicted value",
        ylabel="Fraction of positives",
        title=f"Calibration Curve - {title_suffix}",
    )
    ax.set_xlim(left=-0.05, right=1.05)
    ax.set_ylim(bottom=-0.05, top=1.05)
    ax.legend(loc="lower right")
    return fig


def get_sensitivity_of_top_q_pctl_thresh(
    y_true: pd.Series | Sequence,
    risk_score: pd.Series | Sequence,
    q: float,
    pos_label: PosLabelType,
) -> float:
    """
    Report sensitivity (AKA recall score) using some percentile threshold.

    Calculation:
        number of true positivies / (number of true positives + number of false negatives)
        OR
        number of true positives / total number of actual true class

    Args:
        y_true: true outcome class, of length n_samples
        risk_score: predicted risk scores, of length n_samples
        q: probability for the quantiles to compute
        pos_label: label identifying the positive class in y_true
    """
    if not isinstance(y_true, pd.Series):
        y_true = pd.Series(y_true)
    if not isinstance(risk_score, pd.Series):
        risk_score = pd.Series(risk_score)

    prob_thresh = np.quantile(risk_score, q)
    high_risk = risk_score >= prob_thresh

    # convert actual outcome to booleans to match high_risk array
    y_true = y_true.apply(lambda x: True if x == pos_label else False)

    result = recall_score(y_true, high_risk)
    assert isinstance(result, float)
    return result


def plot_sla_curve(
    y_true: pd.Series,
    risk_score: pd.Series,
    keys: str | list[str],
    title_suffix: str,
    pos_label: PosLabelType,
    alert_rates: list[float] = [0, 0.01, 0.02, 0.03, 0.04, 0.05, 0.06],
    label_alert_rate: float = 0.01,
) -> matplotlib.figure.Figure:
    """
    Create Sensitivity at Low Alert Rates plot

    Args:
        y_true (array-like of shape (n_samples,) or (n_groups,)): overall or group-level true outcome class
        risk_score (array-like of shape (n_samples,) or (n_groups,)): overall or group level predicted risk scores
        keys: overall or subgroup level labels for labeling lines
        title_suffix: suffix for plot title
        pos_label: label identifying the positive class in y_true
        alert_rates: alert rates to plot. Defaults to [0, 0.01, 0.02, 0.03, 0.04, 0.05, 0.06].
        label_alert_rate: alert rate of interest to report sensitivity at. Defaults to 0.01.

    Returns:
        line plot of sensitivity at small alert rates
    """
    if not check_array_of_arrays(y_true):
        y_true = [y_true]
        risk_score = [risk_score]
        keys = [keys]  # type: ignore

    fig, ax = plt.subplots()

    for j in range(len(risk_score)):
        ss = []
        for i in alert_rates:  # calculate sensitivity at different alert rates
            s = get_sensitivity_of_top_q_pctl_thresh(
                y_true[j], risk_score[j], 1 - i, pos_label
            )
            ss.append(s)
        s_lab = round(
            get_sensitivity_of_top_q_pctl_thresh(
                y_true[j], risk_score[j], 1 - label_alert_rate, pos_label
            ),
            2,
        )
        ax.plot(
            alert_rates,
            ss,
            color=PALETTE[j + 1],
            label="{} (Sensitivity at {}% alert rate={})".format(
                keys[j], label_alert_rate * 100, s_lab
            ),
        )

    ax.set(
        xlabel="Alert rate",
        ylabel="sensitivity (true positive rate)",
        title=f"Sensitivity vs. Low Alert Rate - {title_suffix}",
    )
    ax.set_ylim(bottom=-0.02, top=1.02)
    ax.legend(loc="lower right")

    return fig


def compare_trained_models(
    automl_experiment_id: str, automl_metric: str
) -> tuple[pd.DataFrame, str]:
    """
    Retrieve, aggregate and sort performance data for models trained in a specified AutoML experiment.
    The validation dataset is used to tune hyperparameters. Metrics on the validation dataset are used to rank models, so we also use this metric to compare across models.

    Args:
        automl_experiment_id: Experiment ID of the AutoML experiment
        automl_metric: Chosen AutoML optimization metric

    Returns:
        DataFrame containing model types and highest scores for the given metric.
    """
    runs = mlflow.search_runs(
        experiment_ids=[automl_experiment_id], output_format="pandas"
    )
    assert isinstance(runs, pd.DataFrame)  # type guard
    metric = str.lower(automl_metric)
    metric_score_column = (
        f"metrics.val_{metric}"
        if metric == "log_loss"
        else f"metrics.val_{metric}_score"
    )

    runs = runs[["tags.model_type", metric_score_column]].dropna()

    ascending = metric == "log_loss"
    df_sorted = (
        runs.groupby("tags.model_type", as_index=False)
        .agg({metric_score_column: ("min" if ascending else "max")})
        .sort_values(by=metric_score_column, ascending=ascending)
    )
    return df_sorted, metric_score_column


def compare_trained_models_plot(
    automl_experiment_id: str, automl_metric: str
) -> matplotlib.figure.Figure:
    """
    Create a plot to evaluate all the models trained by AutoML.

    Args:
        automl_experiment_id: Experiment ID of the AutoML experiment
        automl_metric: Chosen AutoML optimization metric

    Returns:
        bar chart of model performance on test data by optimization metric.
    """
    df_sorted, metric_score_column = compare_trained_models(
        automl_experiment_id, automl_metric
    )

    df_sorted = df_sorted.sort_values(
        by=metric_score_column,
        ascending=True if not automl_metric == "log_loss" else False,
    )
    fig, ax = plt.subplots()

    ax.set_xlim(
        0,
        (
            df_sorted[metric_score_column].max() + 0.2
            if automl_metric == "log_loss"
            else 1
        ),
    )  # setting buffer after max log_loss to 0.2, for better viz

    colors = ["#A9A9A9"] * len(df_sorted)
    colors[-1] = "#5bc0de"  # Bright blue color for the best model

    bars = ax.barh(
        df_sorted["tags.model_type"], df_sorted[metric_score_column], color=colors
    )

    for bar in bars:
        ax.text(
            bar.get_width()
            - (
                max(df_sorted[metric_score_column]) * 0.005
            ),  # Position the text inside the bar
            bar.get_y() + bar.get_height() / 2,
            f"{bar.get_width():.4f}",
            va="center",
            ha="right",
            fontweight="bold",
            color="white",
            fontsize=12,
        )

    sort_order = (
        "Lowest to Highest" if automl_metric == "log_loss" else "Highest to Lowest"
    )
    automl_metric = (
        automl_metric
        if automl_metric == "log_loss"
        else f"{automl_metric.capitalize()} Score"
    )

    ax.set(
        title=f"{automl_metric} by Model Type {sort_order}",
        facecolor="none",
        frame_on=False,
    )
    ax.tick_params(axis="y", which="both", left=False, right=False, labelleft=True)
    ax.tick_params(axis="x", colors="lightgrey", which="both")  # Color of ticks
    ax.xaxis.grid(True, color="lightgrey", linestyle="--", linewidth=0.5)
    fig.tight_layout()

    return fig


def create_evaluation_plots(
    data: pd.DataFrame,
    risk_score_col: str,
    y_true_col: str,
    pos_label: PosLabelType,
    split_type: str,
) -> tuple[matplotlib.figure.Figure, ...]:
    """
    Create plots to evaluate a model overall - risk score histogram,
    calibration curve, and sensitivity at low alert rates

    Args:
        data: containing predicted and actual outcome data
        risk_score_col: column name containing data of predicted risk scores
        y_true_col: column name containing data of actual outcome classes
        pos_label: label identifying the positive class in y_true
        split_type: type of data being plotted for labeling plots - train, test, or validation

    Returns:
        risk score histogram, calibration curve, and sensitivity at low alert rates figures
    """
    title_suffix = f"{split_type} data - Overall"
    hist_fig = create_risk_score_histogram(data[risk_score_col], title_suffix)
    cal_fig = create_calibration_curve(
        data[y_true_col], data[risk_score_col], "Overall", title_suffix, pos_label
    )
    sla_fig = plot_sla_curve(
        data[y_true_col], data[risk_score_col], "Overall", title_suffix, pos_label
    )
    return hist_fig, cal_fig, sla_fig


def create_evaluation_plots_by_subgroup(
    data: pd.DataFrame,
    risk_score_col: str,
    y_true_col: str,
    pos_label: PosLabelType,
    group_col: str,
    split_type: str,
) -> tuple[matplotlib.figure.Figure, ...]:
    """
    Create plots to evaluate a model by group - calibration curve
    and sensitivity at low alert rates

    Args:
        data: containing predicted and actual outcome data, as well as group label
        risk_score_col: column name containing data of predicted risk scores
        y_true_col: column name containing data of actual outcome classes
        pos_label: label identifying the positive class in y_true
        group_col: column name containing data of subgroup labels
        split_type: type of data being plotted for labeling plots - train, test, or validation

    Returns:
        calibration curve sensitivity at low alert rates figures by group
    """
    title_suffix = f"{split_type} data - {group_col}"

    grouped_data = data.groupby(group_col).agg(list)

    ys = grouped_data[y_true_col]
    scores = grouped_data[risk_score_col]

    subgroups = grouped_data.index
    n_counts = [len(subgroup_scores) for subgroup_scores in scores]
    names = [
        subgroup_name + f" (N={n})" for subgroup_name, n in zip(subgroups, n_counts)
    ]

    if len(subgroups) > 4:
        lowess_frac = 0.7
    else:
        lowess_frac = 0.6

    sla_subgroup_plot = plot_sla_curve(ys, scores, names, title_suffix, pos_label)
    cal_subgroup_plot = create_calibration_curve(
        ys, scores, names, title_suffix, pos_label, lowess_frac=lowess_frac
    )

    return cal_subgroup_plot, sla_subgroup_plot



# Databricks notebook source
# MAGIC %md
# MAGIC # SST Prepare Modeling Dataset: [SCHOOL]
# MAGIC
# MAGIC Second step in the process of transforming raw (PDP) data into actionable, data-driven insights for advisors: featurize the raw, validated data; configure and compute the target variable; perform feature selection; prepare train/test/validation splits; and inspect feature-target associations.
# MAGIC
# MAGIC ### References
# MAGIC
# MAGIC - [Data science product components (Confluence doc)](https://datakind.atlassian.net/wiki/spaces/TT/pages/237862913/Data+science+product+components+the+modeling+process)
# MAGIC - [Databricks Data preparation for classification](https://docs.databricks.com/en/machine-learning/automl/classification-data-prep.html)
# MAGIC - [Databricks runtimes release notes](https://docs.databricks.com/en/release-notes/runtime/index.html)
# MAGIC - [SCHOOL WEBSITE](https://example.com)

# COMMAND ----------

# MAGIC %md
# MAGIC # setup

# COMMAND ----------

# MAGIC %sh python --version

# COMMAND ----------

# install dependencies, most of which should come through our 1st-party SST package
# %pip install git+https://github.com/datakind/student-success-tool.git@develop

# COMMAND ----------

# MAGIC %restart_python

# COMMAND ----------

import logging
import sys

import numpy as np
import pandas as pd
import seaborn as sb
from databricks.connect import DatabricksSession

from student_success_tool.analysis import pdp

# COMMAND ----------

logging.basicConfig(level=logging.INFO)
logging.getLogger("py4j").setLevel(logging.WARNING)  # ignore databricks logger

try:
    spark_session = DatabricksSession.builder.getOrCreate()
except Exception:
    logging.warning("unable to create spark session; are you in a Databricks runtime?")
    pass

# COMMAND ----------

# MAGIC %md
# MAGIC ## `student-success-intervention` hacks

# COMMAND ----------

# MAGIC %load_ext autoreload
# MAGIC %autoreload 2

# COMMAND ----------

# HACK: insert our 1st-party (school-specific) code into PATH
if "../" not in sys.path:
    sys.path.insert(1, "../")

# TODO: specify school's subpackage
from analysis import *  # noqa: F403

# COMMAND ----------

# MAGIC %md
# MAGIC ## Unity Catalog config

# COMMAND ----------

catalog = "sst_dev"

# configure where data is to be read from / written to
inst_name = "SCHOOL"  # TODO: fill in school's name in Unity Catalog
schema = f"{inst_name}_silver"
catalog_schema = f"{catalog}.{schema}"
print(f"{catalog_schema=}")

# COMMAND ----------

# MAGIC %md
# MAGIC # read validated data

# COMMAND ----------

# MAGIC %md
# MAGIC **Note:** Whether you need to use the base or school-specific data schema was determined in the previous step, i.e. the data assessment / EDA notebook.

# COMMAND ----------

df_course = pdp.schemas.RawPDPCourseDataSchema(
    pdp.dataio.read_data_from_delta_table(
        f"{catalog_schema}.course_dataset_validated", spark_session=spark_session
    )
)
print(f"rows x cols = {df_course.shape}")
df_course.head()

# COMMAND ----------

df_cohort = pdp.schemas.RawPDPCohortDataSchema(
    pdp.dataio.read_data_from_delta_table(
        f"{catalog_schema}.cohort_dataset_validated", spark_session=spark_session
    )
)
print(f"rows x cols = {df_cohort.shape}")
df_cohort.head()

# COMMAND ----------

# MAGIC %md
# MAGIC # transform and join datasets

# COMMAND ----------

# school-specific parameters that configure featurization code
# these should all be changed, as desired, and confirmed with client stakeholders
min_passing_grade = pdp.constants.DEFAULT_MIN_PASSING_GRADE
min_num_credits_full_time = pdp.constants.DEFAULT_MIN_NUM_CREDITS_FULL_TIME
course_level_pattern = pdp.constants.DEFAULT_COURSE_LEVEL_PATTERN
key_course_subject_areas = None
key_course_ids = None

# COMMAND ----------

df_student_terms = pdp.dataops.make_student_term_dataset(
    df_cohort,
    df_course,
    min_passing_grade=min_passing_grade,
    min_num_credits_full_time=min_num_credits_full_time,
    course_level_pattern=course_level_pattern,
    key_course_subject_areas=key_course_subject_areas,
    key_course_ids=key_course_ids,
)
df_student_terms

# COMMAND ----------

# take a peek at featurized columns -- it's a lot
df_student_terms.columns.tolist()

# COMMAND ----------

# save student-term dataset in unity catalog, as needed
# write_table_path = f"{catalog_schema}.student_term_dataset"
# pdp.dataio.write_data_to_delta_table(df_student_terms, write_table_path, spark_session)

# COMMAND ----------

# MAGIC %md
# MAGIC # filter students and compute target

# COMMAND ----------

# MAGIC %md
# MAGIC A single function call can be used to make a labeled dataset for modeling, which includes selecting eligible students, subsetting to just the terms from which predictions are made, and computing target variables. _Which_ function depends primarily on the target to be computed: a failure to earn enough credits within a timeframe since initial enrollment, or a particular mid-way checkpoint (other targets pending). Input parameters will vary depending on the school and the target.
# MAGIC
# MAGIC For example, here's how one could make a labeled dataset with the following setup: Filters for first-time students, enrolled either full- or part-time, seeking an Associate's degree; that earn at least 60 credits within 3 years of enrollment if full-time or 6 years of enrollment if part-time; making predictions from the first term for which they've earned 30 credits.
# MAGIC
# MAGIC ```python
# MAGIC # school-specific parameters that configure target variable code
# MAGIC min_num_credits_checkin = 30.0
# MAGIC min_num_credits_target = 60.0
# MAGIC student_criteria = {
# MAGIC     "enrollment_type": "FIRST-TIME",
# MAGIC     "enrollment_intensity_first_term": ["FULL-TIME", "PART-TIME"],
# MAGIC     "credential_type_sought_year_1": "Associate's Degree",
# MAGIC }
# MAGIC intensity_time_limits = [
# MAGIC     ("FULL-TIME", 3.0, "year"),
# MAGIC     ("PART-TIME", 6.0, "year"),
# MAGIC ]
# MAGIC
# MAGIC df_labeled = pdp.targets.failure_to_earn_enough_credits_in_time_from_enrollment.make_labeled_dataset(
# MAGIC     df_student_terms,
# MAGIC     min_num_credits_checkin=min_num_credits_checkin,
# MAGIC     min_num_credits_target=min_num_credits_target,
# MAGIC     student_criteria=student_criteria,
# MAGIC     intensity_time_limits=intensity_time_limits,
# MAGIC )
# MAGIC ```
# MAGIC
# MAGIC Check out the `pdp.targets` module for options and more info.

# COMMAND ----------

# TODO: school-specific parameters that configure target variable code

# COMMAND ----------

# TODO: suitable function call for school's use case
# df_labeled = pdp.targets.*.make_labeled_dataset(...)
df_labeled = pd.DataFrame()

# COMMAND ----------

ax = sb.histplot(
    df_labeled.sort_values("cohort"),
    y="cohort",
    multiple="stack",
    shrink=0.75,
    edgecolor="white",
)
_ = ax.set(xlabel="Number of Students")

# COMMAND ----------

ax = sb.histplot(
    df_labeled.sort_values("cohort"),
    y="cohort",
    hue="target",
    multiple="stack",
    shrink=0.75,
    edgecolor="white",
)
_ = ax.set(xlabel="Number of Students")

# COMMAND ----------

# drop unwanted columns and mask values by time
df_labeled = pdp.dataops.clean_up_labeled_dataset_cols_and_vals(df_labeled)
df_labeled.shape

# COMMAND ----------

# save labeled dataset in unity catalog, as needed
# write_table_path = f"{catalog}.{schema}.labeled_dataset"
# pdp.dataio.write_data_to_delta_table(df_labeled, write_table_path, spark_session=spark_session)

# COMMAND ----------

# MAGIC %md
# MAGIC # feature selection + splits

# COMMAND ----------

import mlflow

from student_success_tool.modeling.feature_selection import select_features

# databricks freaks out during feature selection if autologging isn't disabled
mlflow.autolog(disable=True)

# COMMAND ----------

non_feature_cols = [
    "student_guid",
    "target",
    "student_age",
    "race",
    "ethnicity",
    "gender",
    "first_gen",
]
df_labeled_selected = select_features(
    df_labeled,
    non_feature_cols=non_feature_cols,
    # TODO: configure these params as desired
    force_include_cols=[],
    incomplete_threshold=0.5,
    low_variance_threshold=0.0,
    collinear_threshold=10.0,
)
df_labeled_selected

# COMMAND ----------

# heads-up: AutoML doesn't preserve Student IDs in the training data!
# manually assigning splits lets us know which rows were in which partition
# when evaluating the model across student groups in the validation set
# default split strategy by AutoML is [0.6, 0.2, 0.2] for train/test/validation
np.random.seed(1)
df_labeled_selected = df_labeled_selected.assign(
    split=lambda df: np.random.choice(
        ["train", "test", "validate"],
        size=df.shape[0],
        p=[0.6, 0.2, 0.2],
    )
)
df_labeled_selected["split"].value_counts(normalize=True)

# COMMAND ----------

# save final modeling dataset in unity catalog, as needed
# write_table_path = f"{catalog}.{schema}.modeling_dataset"
# pdp.dataio.write_data_to_delta_table(df_labeled_selected, write_table_path, spark_session=spark_session)

# COMMAND ----------

# MAGIC %md
# MAGIC # feature-target associations

# COMMAND ----------

target_assocs = pdp.eda.compute_pairwise_associations(
    df_labeled_selected, ref_col="target"
)
target_assocs

# COMMAND ----------

# MAGIC %md
# MAGIC # Wrap-up
# MAGIC
# MAGIC TODO

# COMMAND ----------



# ruff: noqa: F821
# mypy: ignore-errors
import functools as ft
import typing as t

import pandas as pd
import pandera as pda
import pandera.typing as pt

StudentAgeField = ft.partial(
    pda.Field,
    dtype_kwargs={
        "categories": ["20 AND YOUNGER", ">20 - 24", "OLDER THAN 24"],
        "ordered": True,
    },
)
RaceField = ft.partial(
    pda.Field,
    dtype_kwargs={
        "categories": [
            "NONRESIDENT ALIEN",
            "AMERICAN INDIAN OR ALASKA NATIVE",
            "ASIAN",
            "BLACK OR AFRICAN AMERICAN",
            "NATIVE HAWAIIAN OR OTHER PACIFIC ISLANDER",
            "WHITE",
            "HISPANIC",
            "TWO OR MORE RACES",
            "UNKNOWN",
        ],
    },
)
EthnicityField = ft.partial(pda.Field, dtype_kwargs={"categories": ["H", "N", "UK"]})
GenderField = ft.partial(
    pda.Field, dtype_kwargs={"categories": ["M", "F", "P", "X", "UK"]}
)
TermField = ft.partial(
    pda.Field,
    dtype_kwargs={
        "categories": ["FALL", "WINTER", "SPRING", "SUMMER"],
        "ordered": True,
    },
)
GPAField = ft.partial(pda.Field, nullable=True, ge=0.0, le=4.0)
NumCreditsGt0Field = ft.partial(pda.Field, nullable=True, ge=0.0)
GradeField = ft.partial(pda.Field, nullable=True)
CompletedDevOrGatewayField = ft.partial(
    pda.Field, nullable=True, dtype_kwargs={"categories": ["C", "D", "NA"]}
)
YearsToOfField = ft.partial(pda.Field, ge=0, le=8)
LocaleField = ft.partial(
    pda.Field,
    nullable=True,
    dtype_kwargs={"categories": ["URBAN", "SUBURB", "TOWN/RURAL"]},
)


class RawPDPCohortDataSchema(pda.DataFrameModel):
    """
    Schema (aka ``DataFrameModel``) for raw PDP cohort data that validates columns,
    data types (including categorical categories), acceptable value ranges, and more.

    References:
        - https://help.studentclearinghouse.org/pdp/knowledge-base/cohort-level-analysis-ready-file-data-dictionary/
        - https://pandera.readthedocs.io/en/stable/dataframe_models.html
    """

    student_guid: pt.Series["string"]
    institution_id: pt.Series["string"]
    cohort: pt.Series["string"]
    cohort_term: pt.Series[pd.CategoricalDtype] = TermField()
    student_age: pt.Series[pd.CategoricalDtype] = StudentAgeField()
    enrollment_type: pt.Series[pd.CategoricalDtype] = pda.Field(
        dtype_kwargs={"categories": ["FIRST-TIME", "RE-ADMIT", "TRANSFER-IN"]},
    )
    # NOTE: categories set in a parser, which forces "UK" / "UNKNOWN" values to null
    enrollment_intensity_first_term: pt.Series[pd.CategoricalDtype] = pda.Field()
    # NOTE: categories set in a parser, which forces "UK" values to null
    math_placement: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    english_placement: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    dual_and_summer_enrollment: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True
    )
    race: pt.Series[pd.CategoricalDtype] = RaceField()
    ethnicity: pt.Series[pd.CategoricalDtype] = EthnicityField()
    gender: pt.Series[pd.CategoricalDtype] = GenderField()
    first_gen: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True, dtype_kwargs={"categories": ["P", "C", "A", "B"]}
    )
    # NOTE: categories set in a parser, which forces "UK" values to null
    pell_status_first_year: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    attendance_status_term_1: pt.Series[pd.CategoricalDtype] = pda.Field(
        dtype_kwargs={
            "categories": [
                "First-Time Full-Time",
                "First-Time Part-Time",
                "First-Time Unknown",
                "Transfer-In Full-Time",
                "Transfer-In Part-Time",
                "Transfer-In Unknown",
                "Re-admit Full-Time",
                "Re-admit Part-Time",
                "Re-admit Unknown",
            ]
        },
    )
    credential_type_sought_year_1: pt.Series[pd.CategoricalDtype] = pda.Field(
        dtype_kwargs={
            "categories": [
                "Less than one-year certificate, less than Associate degree",
                "One to two year certificate, less than Associate degree",
                "Two to four year certificate, less than Bachelor's degree",
                "Undergraduate Certificate or Diploma Program",
                "Associate Degree",
                "Bachelor's Degree",
                "Post Baccalaureate Certificate",
                "Master's Degree",
                "Doctoral Degree",
                "First Professional Degree",
                "Graduate/Professional Certificate",
                "Non- Credential Program (Preparatory Coursework/Teach Certification)",
                "Missing",
            ]
        },
    )
    program_of_study_term_1: pt.Series["string"] = pda.Field(nullable=True)
    gpa_group_term_1: pt.Series["Float32"] = GPAField()
    gpa_group_year_1: pt.Series["Float32"] = GPAField()
    number_of_credits_attempted_year_1: pt.Series["Float32"] = pda.Field(
        nullable=True, ge=1.0, raise_warning=True
    )
    number_of_credits_earned_year_1: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_2: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_2: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_3: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_3: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_4: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_4: pt.Series["Float32"] = NumCreditsGt0Field()
    # NOTE: categories set in a parser, which forces "UK" values to null
    gateway_math_status: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    gateway_english_status: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    attempted_gateway_math_year_1: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True
    )
    # NOTE: categories set in a parser, which forces "UK" values to null
    attempted_gateway_english_year_1: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True
    )
    completed_gateway_math_year_1: pt.Series[pd.CategoricalDtype] = (
        CompletedDevOrGatewayField()
    )
    completed_gateway_english_year_1: pt.Series[pd.CategoricalDtype] = (
        CompletedDevOrGatewayField()
    )
    gateway_math_grade_y_1: pt.Series["string"] = GradeField()
    gateway_english_grade_y_1: pt.Series["string"] = GradeField()
    # NOTE: categories set in a parser, which forces "UK" values to null
    attempted_dev_math_y_1: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    attempted_dev_english_y_1: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    completed_dev_math_y_1: pt.Series[pd.CategoricalDtype] = (
        CompletedDevOrGatewayField()
    )
    completed_dev_english_y_1: pt.Series[pd.CategoricalDtype] = (
        CompletedDevOrGatewayField()
    )
    retention: pt.Series["boolean"]
    persistence: pt.Series["boolean"]
    years_to_bachelors_at_cohort_inst: pt.Series["Int8"] = YearsToOfField(nullable=True)
    years_to_associates_or_certificate_at_cohort_inst: pt.Series["Int8"] = (
        YearsToOfField(nullable=True)
    )
    years_to_bachelor_at_other_inst: pt.Series["Int8"] = YearsToOfField(nullable=True)
    years_to_associates_or_certificate_at_other_inst: pt.Series["Int8"] = (
        YearsToOfField(nullable=True)
    )
    years_of_last_enrollment_at_cohort_institution: pt.Series["int32"] = (
        YearsToOfField()
    )
    years_of_last_enrollment_at_other_institution: pt.Series["int32"] = YearsToOfField()
    time_to_credential: pt.Series["Float32"] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which assigns "UK" values to null
    reading_placement: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True
    )
    special_program: t.Optional[pt.Series["string"]] = pda.Field(nullable=True)
    naspa_first_generation: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True,
        dtype_kwargs={"categories": ["-1", "0", "1", "2", "3", "4", "5", "6"]},
    )
    # NOTE: categories set in a parser, which assigns "UK" values to null
    incarcerated_status: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True
    )
    # NOTE: categories set in a parser, which forces "-1" / "-1.0" values to null
    military_status: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True
    )
    employment_status: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True, dtype_kwargs={"categories": ["-1", "0", "1", "2", "3", "4"]}
    )
    # NOTE: categories set in a parser, which assigns "UK" values to null
    disability_status: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True
    )
    foreign_language_completion: t.Optional[pt.Series["string"]] = pda.Field(
        nullable=True
    )
    first_year_to_bachelors_at_cohort_inst: pt.Series["Int8"] = YearsToOfField(
        nullable=True
    )
    first_year_to_associates_or_certificate_at_cohort_inst: pt.Series["Int8"] = (
        YearsToOfField(nullable=True)
    )
    first_year_to_bachelor_at_other_inst: pt.Series["Int8"] = YearsToOfField(
        nullable=True
    )
    first_year_to_associates_or_certificate_at_other_inst: pt.Series["Int8"] = (
        YearsToOfField(nullable=True)
    )
    program_of_study_year_1: pt.Series["string"] = pda.Field(nullable=True)
    most_recent_bachelors_at_other_institution_state: pt.Series["string"] = pda.Field(
        nullable=True
    )
    most_recent_associates_or_certificate_at_other_institution_state: pt.Series[
        "string"
    ] = pda.Field(nullable=True)
    most_recent_last_enrollment_at_other_institution_state: pt.Series["string"] = (
        pda.Field(nullable=True)
    )
    first_bachelors_at_other_institution_state: pt.Series["string"] = pda.Field(
        nullable=True
    )
    first_associates_or_certificate_at_other_institution_state: pt.Series["string"] = (
        pda.Field(nullable=True)
    )
    most_recent_bachelors_at_other_institution_carnegie: pt.Series["string"] = (
        pda.Field(nullable=True)
    )
    most_recent_associates_or_certificate_at_other_institution_carnegie: pt.Series[
        "string"
    ] = pda.Field(nullable=True)
    most_recent_last_enrollment_at_other_institution_carnegie: pt.Series["string"] = (
        pda.Field(nullable=True)
    )
    first_bachelors_at_other_institution_carnegie: pt.Series["string"] = pda.Field(
        nullable=True
    )
    first_associates_or_certificate_at_other_institution_carnegie: pt.Series[
        "string"
    ] = pda.Field(nullable=True)
    most_recent_bachelors_at_other_institution_locale: pt.Series[
        pd.CategoricalDtype
    ] = LocaleField()
    most_recent_associates_or_certificate_at_other_institution_locale: pt.Series[
        pd.CategoricalDtype
    ] = LocaleField()
    most_recent_last_enrollment_at_other_institution_locale: pt.Series[
        pd.CategoricalDtype
    ] = LocaleField()
    first_bachelors_at_other_institution_locale: pt.Series[pd.CategoricalDtype] = (
        LocaleField()
    )
    first_associates_or_certificate_at_other_institution_locale: pt.Series[
        pd.CategoricalDtype
    ] = LocaleField()

    @pda.parser("enrollment_intensity_first_term")
    def set_enrollment_intensity_first_term_categories(cls, series):
        return series.cat.set_categories(["FULL-TIME", "PART-TIME"])

    @pda.parser("math_placement", "english_placement", "reading_placement")
    def set_subj_placement_categories(cls, series):
        return series.cat.set_categories(["C", "N"])

    @pda.parser("dual_and_summer_enrollment")
    def set_dual_and_summer_enrollment_categories(cls, series):
        return series.cat.set_categories(["DE", "SE", "DS"])

    @pda.parser("pell_status_first_year")
    def set_pell_status_first_year_categories(cls, series):
        return series.cat.set_categories(["Y", "N"])

    @pda.parser("gateway_math_status", "gateway_english_status")
    def set_gateway_math_english_status_categories(cls, series):
        return series.cat.set_categories(["R", "N"])

    @pda.parser("attempted_gateway_math_year_1", "attempted_gateway_english_year_1")
    def set_attempted_gateway_math_english_year_1_categories(cls, series):
        return series.cat.set_categories(["Y", "N"])

    @pda.parser("attempted_dev_math_y_1", "attempted_dev_english_y_1")
    def set_attempted_dev_math_english_y_1_categories(cls, series):
        return series.cat.set_categories(["Y", "N", "NA"])

    @pda.parser("incarcerated_status")
    def set_incarcerated_status_categories(cls, series):
        return series.cat.set_categories(["Y", "P", "N"])

    @pda.parser("disability_status")
    def set_disability_status_categories(cls, series):
        return series.cat.set_categories(["Y", "N"])

    @pda.parser("military_status")
    def set_military_status_categories(cls, series):
        return (
            series.astype("Float32")
            .astype("Int8")
            .astype("category")
            .cat.set_categories(["0", "1", "2"])
        )

    @pda.parser(
        "years_to_associates_or_certificate_at_cohort_inst",
        "years_to_bachelors_at_cohort_inst",
        "years_to_associates_or_certificate_at_other_inst",
        "years_to_bachelor_at_other_inst",
        "first_year_to_associates_or_certificate_at_cohort_inst",
        "first_year_to_bachelors_at_cohort_inst",
        "first_year_to_associates_or_certificate_at_other_inst",
        "first_year_to_bachelor_at_other_inst",
    )
    def set_zero_year_values_to_null(cls, series):
        return series.mask(series.eq(0), pd.NA).astype("Int8")

    @pda.check("institution_id", name="check_num_institutions")
    def check_num_institutions(cls, series) -> bool:
        return series.nunique() == 1

    class Config:
        coerce = True
        strict = True
        unique_column_names = True
        add_missing_columns = False
        drop_invalid_rows = False
        unique = ["student_guid"]



import numpy as np
import pandas as pd
import pytest

from student_success_tool.analysis.pdp import dataops


@pytest.mark.parametrize(
    ["series", "exp"],
    [
        (
            pd.Series(
                ["WI", "SP", "FA", "SU"],
                dtype=pd.CategoricalDtype(["FA", "WI", "SP", "SU"], ordered=True),
            ),
            "FA",
        ),
        (
            pd.Series(
                ["FA", "SU", "SP"],
                dtype=pd.CategoricalDtype(["SU", "FA", "SP"], ordered=True),
            ),
            "SU",
        ),
    ],
)
def test_infer_first_term_of_year(series, exp):
    obs = dataops.infer_first_term_of_year(series)
    assert obs == exp


@pytest.mark.parametrize(
    ["series", "exp"],
    [
        (pd.Series(["WI", "SP", "FA", "SU"], dtype="category"), 4),
        (pd.Series(["SU", "FA", "WI"], dtype="category"), 3),
    ],
)
def test_infer_num_terms_in_year(series, exp):
    obs = dataops.infer_num_terms_in_year(series)
    assert obs == exp


@pytest.mark.parametrize(
    ["df", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "cohort": ["2024-25", "2023-24"],
                    "cohort_term": ["FALL", "SPRING"],
                    "time_to_credential": [2.5, 3.0],
                    "attendance_status_term_1": [
                        "FIRST-TIME FULL-TIME",
                        "TRANSFER-IN PART-TIME",
                    ],
                    "number_of_credits_attempted_year_1": [10.0, 5.0],
                    "number_of_credits_earned_year_1": [10.0, 5.0],
                    "gateway_math_status": ["R", "N"],
                    "gpa_group_year_1": [3.0, 4.0],
                    "most_recent_bachelors_at_other_institution_state": ["VT", "IL"],
                }
            ),
            pd.DataFrame(
                {
                    "cohort": ["2024-25", "2023-24"],
                    "cohort_term": ["FALL", "SPRING"],
                    "number_of_credits_attempted_year_1": [10.0, 5.0],
                    "number_of_credits_earned_year_1": [10.0, 5.0],
                    "gpa_group_year_1": [3.0, 4.0],
                }
            ),
        ),
    ],
)
def test_standardize_cohort_dataset(df, exp):
    obs = dataops.standardize_cohort_dataset(df)
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "academic_year": ["2020-21", "2020-21", "2021-22"],
                    "academic_term": ["FALL", "SPRING", "FALL"],
                    "course_prefix": ["MATH", "MATH", "PHYS"],
                    "course_number": ["101", "202", "303"],
                    "course_name": ["NAME1", "NAME2", "NAME3"],
                    "course_type": ["CU", "CU", "CC"],
                    "grade": ["4", "1", "W"],
                    "core_course": ["Y", "N", "Y"],
                    "core_course_type": ["TYPE1", "TYPE2", "TYPE3"],
                    "core_competency_completed": ["Y", None, None],
                    "cohort": ["2020-21", "2020-21", "2020-21"],
                    "cohort_term": ["FALL", "FALL", "SPRING"],
                    "student_age": ["20 AND YOUNGER", ">20 - 24", "OLDER THAN 24"],
                    "race": ["WHITE", "HISPANIC", "ASIAN"],
                    "ethnicity": ["N", "H", "N"],
                    "gender": ["M", "F", "X"],
                    "credential_engine_identifier": [None, None, None],
                    "enrollment_record_at_other_institution_s_state_s": [
                        "VT",
                        None,
                        None,
                    ],
                    "enrollment_record_at_other_institution_s_carnegie_s": [
                        "Doctoral Universities",
                        None,
                        None,
                    ],
                    "enrollment_record_at_other_institution_s_locale_s": [
                        "Urban",
                        None,
                        None,
                    ],
                }
            ),
            pd.DataFrame(
                {
                    "academic_year": ["2020-21", "2020-21", "2021-22"],
                    "academic_term": ["FALL", "SPRING", "FALL"],
                    "course_prefix": ["MATH", "MATH", "PHYS"],
                    "course_number": ["101", "202", "303"],
                    "course_type": ["CU", "CU", "CC"],
                    "grade": ["4", "1", "W"],
                    "core_course": ["Y", "N", "Y"],
                }
            ),
        ),
    ],
)
def test_standardize_course_dataset(df, exp):
    obs = dataops.standardize_course_dataset(df)
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "year_of_enrollment_at_cohort_inst": [1, 2, 3, 4],
                    "cumnum_terms_enrolled": [2, 3, 4, 5],
                    "term_id": [
                        "2020-21 FALL",
                        "2020-21 WINTER",
                        "2020-21 SPRING",
                        "2020-21 SUMMER",
                    ],
                    "course_ids": [
                        ["X101", "Y101"],
                        ["X102", "Y101", "Z201"],
                        ["X101", "Y102"],
                        ["X101", "Y101", "X102", "Y202", "Z101"],
                    ],
                    "retention": [False, True, True, True],
                    "first_year_to_associates_or_certificate_at_cohort_inst": [
                        pd.NA,
                        2,
                        2,
                        5,
                    ],
                    "first_year_to_bachelors_at_cohort_inst": [pd.NA, pd.NA, 4, 3],
                    "first_year_to_associates_or_certificate_at_other_inst": [
                        4,
                        1,
                        5,
                        3,
                    ],
                    "first_year_to_bachelor_at_other_inst": [pd.NA, pd.NA, pd.NA, 6],
                    "frac_credits_earned_year_1": [1.0, 0.5, 0.75, 0.9],
                    "frac_credits_earned_year_2": [0.9, 0.75, 0.8, 0.85],
                    "num_courses_diff_term_2_to_term_3": [0.0, 1.0, -1.0, 0.0],
                    "num_courses_diff_term_3_to_term_4": [1.0, -1.0, 0.0, 1.0],
                }
            ).astype(
                {
                    "first_year_to_associates_or_certificate_at_cohort_inst": "Int8",
                    "first_year_to_bachelors_at_cohort_inst": "Int8",
                    "first_year_to_associates_or_certificate_at_other_inst": "Int8",
                    "first_year_to_bachelor_at_other_inst": "Int8",
                }
            ),
            pd.DataFrame(
                {
                    "year_of_enrollment_at_cohort_inst": [1, 2, 3, 4],
                    "cumnum_terms_enrolled": [2, 3, 4, 5],
                    "first_year_to_associates_or_certificate_at_cohort_inst": [
                        pd.NA,
                        pd.NA,
                        2,
                        pd.NA,
                    ],
                    "first_year_to_bachelors_at_cohort_inst": [pd.NA, pd.NA, pd.NA, 3],
                    "first_year_to_associates_or_certificate_at_other_inst": [
                        pd.NA,
                        1,
                        pd.NA,
                        3,
                    ],
                    "first_year_to_bachelor_at_other_inst": [
                        pd.NA,
                        pd.NA,
                        pd.NA,
                        pd.NA,
                    ],
                    "frac_credits_earned_year_1": [1.0, 0.5, 0.75, 0.9],
                    "frac_credits_earned_year_2": [np.nan, 0.75, 0.8, 0.85],
                    "num_courses_diff_term_2_to_term_3": [np.nan, 1.0, -1.0, 0.0],
                    "num_courses_diff_term_3_to_term_4": [np.nan, np.nan, 0.0, 1.0],
                }
            ).astype(
                {
                    "first_year_to_associates_or_certificate_at_cohort_inst": "Int8",
                    "first_year_to_bachelors_at_cohort_inst": "Int8",
                    "first_year_to_associates_or_certificate_at_other_inst": "Int8",
                    "first_year_to_bachelor_at_other_inst": "Int8",
                }
            ),
        ),
    ],
)
def test_clean_up_labeled_dataset_cols_and_vals(df, exp):
    obs = dataops.clean_up_labeled_dataset_cols_and_vals(df)
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty



import logging
import typing as t
from collections.abc import Collection

import numpy as np
import pandas as pd

from .. import dataops, features, utils
from . import shared

LOGGER = logging.getLogger(__name__)


def make_labeled_dataset(
    df: pd.DataFrame,
    *,
    min_num_credits_checkin: float,
    min_num_credits_target: float,
    student_criteria: dict[str, object | Collection[object]],
    intensity_time_limits: list[tuple[str, float, t.Literal["year", "term"]]],
    student_id_cols: str | list[str] = "student_guid",
    enrollment_intensity_col: str = "enrollment_intensity_first_term",
    num_credits_col: str = "num_credits_earned_cumsum",
    term_col: str = "academic_term",
    term_rank_col: str = "term_rank",
) -> pd.DataFrame:
    """
    Make a labeled dataset for modeling, where each row consists of features
    from eligible students' first qualifying term matched to computed target variables.

    Args:
        df
        min_num_credits_checkin
        min_num_credits_target
        student_criteria
        intensity_time_limits
        student_id_cols
        enrollment_intensity_col
        num_credits_col
        term_col
        term_rank_col

    See Also:
        - :func:`select_eligible_students()`
        - :func:`compute_target_variable()`
        - :func:`shared.get_first_student_terms_at_num_credits_earned()`
    """
    df_eligible_students = select_eligible_students(
        df,
        student_criteria=student_criteria,
        intensity_time_limits=intensity_time_limits,
        min_num_credits_checkin=min_num_credits_checkin,
        student_id_cols=student_id_cols,
        enrollment_intensity_col=enrollment_intensity_col,
        num_credits_col=num_credits_col,
        term_col=term_col,
        term_rank_col=term_rank_col,
    )
    df_eligible_student_terms = pd.merge(
        df, df_eligible_students, on=student_id_cols, how="inner"
    )
    df_features = shared.get_first_student_terms_at_num_credits_earned(
        df_eligible_student_terms,
        min_num_credits=min_num_credits_checkin,
        student_id_cols=student_id_cols,
        include_cols=None,
    )
    df_targets = compute_target_variable(
        df_eligible_student_terms,
        min_num_credits_target=min_num_credits_target,
        intensity_time_limits=intensity_time_limits,
        student_id_cols=student_id_cols,
    )
    df_labeled = pd.merge(df_features, df_targets, on=student_id_cols, how="inner")
    return df_labeled


def compute_target_variable(
    df: pd.DataFrame,
    *,
    min_num_credits_target: float,
    intensity_time_limits: list[tuple[str, float, t.Literal["year", "term"]]],
    student_id_cols: str | list[str] = "student_guid",
    enrollment_intensity_col: str = "enrollment_intensity_first_term",
    num_credits_col: str = "num_credits_earned_cumsum",
    term_col: str = "academic_term",
    term_rank_col: str = "term_rank",
) -> pd.Series:
    """
    Args:
        df: Student-term dataset, with cumulative features added.
        min_num_credits_target: Minimum number of credits earned
            within specified time limits to be considered a *success* case.
        intensity_time_limits
        student_id_cols: Columns that uniquely identify students, used for grouping rows.
        enrollment_intensity_col
        num_credits_col
        term_col
        term_rank_col
    """
    student_id_cols = utils.to_list(student_id_cols)
    include_cols = [enrollment_intensity_col]
    # we want a target for every student in input df; this will ensure it
    df_distinct_students = df[student_id_cols].drop_duplicates(ignore_index=True)
    df_at_start = shared.get_first_student_terms(
        df,
        student_id_cols=student_id_cols,
        sort_cols=term_rank_col,
        include_cols=[enrollment_intensity_col],
    )
    df_at_target = shared.get_first_student_terms_at_num_credits_earned(
        df,
        min_num_credits=min_num_credits_target,
        student_id_cols=student_id_cols,
        include_cols=include_cols,
        sort_cols=term_rank_col,
        num_credits_col=num_credits_col,
    )
    df_at = pd.merge(
        df_at_start,
        df_at_target,
        on=student_id_cols,
        how="left",
        suffixes=("_start", "_target"),
    )
    num_terms_in_year = dataops.infer_num_terms_in_year(df[term_col])
    intensity_num_terms = shared._compute_intensity_num_terms(
        intensity_time_limits, num_terms_in_year
    )
    # compute all intensity/term boolean arrays separately
    # then combine with a logical OR
    trcol = term_rank_col  # hack, so logic below fits on lines
    targets = [
        (
            # enrollment intensity is equal to a specified value
            df_at[f"{enrollment_intensity_col}_start"].eq(intensity)
            & (
                (df_at[f"{trcol}_target"] - df_at[f"{trcol}_start"]).gt(num_terms)
                # or they *never* earned enough credits for target
                | df_at[f"{trcol}_target"].isna()
            )
        )
        for intensity, num_terms in intensity_num_terms
    ]
    target = np.logical_or.reduce(targets)
    # assign True to all students passing intensity/year condition(s) above
    df_target_true = (
        df_at.loc[target, student_id_cols]
        .assign(target=True)
        .astype({"target": "boolean"})
    )
    return (
        # all students not assigned True, now assigned False
        pd.merge(df_distinct_students, df_target_true, on=student_id_cols, how="left")
        .fillna(False)
        .astype({"target": "bool"})
        # TODO: do we want a series with student ids as index and target as values, or nah?
        .set_index(student_id_cols)
        .loc[:, "target"]
    )


def select_eligible_students(
    df: pd.DataFrame,
    *,
    student_criteria: dict[str, object | Collection[object]],
    intensity_time_limits: list[tuple[str, float, t.Literal["year", "term"]]],
    min_num_credits_checkin: float,
    student_id_cols: str | list[str] = "student_guid",
    enrollment_intensity_col: str = "enrollment_intensity_first_term",
    num_credits_col: str = "num_credits_earned_cumsum",
    term_col: str = "academic_term",
    term_rank_col: str = "term_rank",
) -> pd.DataFrame:
    """
    Args:
        df
        student_criteria
        intensity_time_limits
        min_num_credits_checkin
        student_id_cols
        enrollment_intensity_col
        num_credits_col
        term_col
        term_rank_col
    """
    nuq_students_in = df.groupby(by=student_id_cols, sort=False).ngroups
    df_students_by_criteria = shared.select_students_by_criteria(
        df, student_id_cols=student_id_cols, **student_criteria
    )
    max_term_rank = df[term_rank_col].max()
    num_terms_in_year = dataops.infer_num_terms_in_year(df[term_col])
    df_students_by_num_creds = shared.get_first_student_terms_at_num_credits_earned(
        df,
        min_num_credits=min_num_credits_checkin,
        student_id_cols=student_id_cols,
        sort_cols=term_rank_col,
        num_credits_col=num_credits_col,
        include_cols=[],
    ).loc[:, utils.to_list(student_id_cols)]
    nuq_students_checkin = len(df_students_by_num_creds)
    shared._log_eligible_selection(
        nuq_students_in, nuq_students_checkin, "check-in credits earned"
    )
    df_ref = shared.get_first_student_terms(
        df,
        student_id_cols=student_id_cols,
        sort_cols=term_rank_col,
        include_cols=[enrollment_intensity_col],
    )
    df_students_by_time_left = shared.select_students_by_time_left(
        df_ref,
        intensity_time_lefts=intensity_time_limits,
        max_term_rank=max_term_rank,
        num_terms_in_year=num_terms_in_year,
        student_id_cols=student_id_cols,
        enrollment_intensity_col=enrollment_intensity_col,
        term_rank_col=term_rank_col,
    )
    df_out = features.shared.merge_many_dataframes(
        [df_students_by_criteria, df_students_by_num_creds, df_students_by_time_left],
        on=student_id_cols,
        how="inner",
    )
    nuq_students_out = df_out.groupby(by=student_id_cols, sort=False).ngroups
    shared._log_eligible_selection(nuq_students_in, nuq_students_out, "overall")
    return df_out



import numpy as np
import pandas as pd


def select_top_features_for_display(
    features: pd.DataFrame,
    unique_ids: pd.Series,
    predicted_probabilities: list[float],
    shap_values: pd.Series,
    n_features: int = 3,
) -> pd.DataFrame:
    """
    Select most important features from SHAP for each student
    and format for display

    Args:
        features: features used in modeling
        unique_ids: student IDs, of length ``features.shape[0]``
        predicted_probabilities: predicted probabilities for each student, in the same
            order as unique_ids, of shape len(unique_ids)
        shap_values: array of arrays of SHAP values, of shape len(unique_ids)
        n_features: number of important features to return

    Returns:
        explainability dataframe for display
    """
    top_features_info = []

    for i, (unique_id, predicted_proba) in enumerate(
        zip(unique_ids, predicted_probabilities)
    ):
        instance_shap_values = shap_values[i]
        top_indices = np.argsort(-np.abs(instance_shap_values))[:n_features]
        top_features = features.columns[top_indices]
        top_shap_values = instance_shap_values[top_indices]

        for rank, (feature, shap_value) in enumerate(
            zip(top_features, top_shap_values), start=1
        ):
            top_features_info.append(
                {
                    "Student ID": unique_id,
                    "Support Score": predicted_proba,
                    "Top Indicators": feature,
                    "SHAP Value": shap_value,
                    "Rank": rank,
                }  # column names defined here https://app.asana.com/0/1206275396780585/1206834683873668/f
            )
    return pd.DataFrame(top_features_info)



import logging
import typing as t
from collections.abc import Collection

import numpy as np
import pandas as pd

from .. import utils

LOGGER = logging.getLogger(__name__)


def select_students_by_criteria(
    df: pd.DataFrame,
    *,
    student_id_cols: str | list[str] = "student_guid",
    **criteria: object | Collection[object],
) -> pd.DataFrame:
    """
    Select distinct students in ``df`` that are eligible according to one or more criteria.

    Args:
        df: Student-term dataset. For example, as output by
            :func:`pdp.dataops.make_student_term_dataset()` .
        student_id_cols: Column(s) that uniquely identify students in ``df`` ,
            used to drop duplicates.
        **criteria: Column name in ``df`` mapped to one or multiple values
            that it must equal in order for the corresponding student to be considered
            "eligible". Multiple criteria are combined with a logical "AND". For example,
            ``enrollment_type="FIRST-TIME", enrollment_intensity_first_term=["FULL-TIME", "PART-TIME"]``
            selects first-time students who enroll at full- or part-time intensity.

    Warning:
        This assumes that students' eligibility criteria values are the same across all
        terms. If that assumption is violated, this code actually checks if the student
        satisfied all of the criteria during _any_ of their terms.
    """
    if not criteria:
        raise ValueError("one or more eligibility criteria must be specified")

    student_id_cols = utils.to_list(student_id_cols)
    nuq_students_in = df.groupby(by=student_id_cols, sort=False).ngroups
    is_eligibles = [
        df[key].isin(set(val))  # type: ignore
        if utils.is_collection_but_not_string(val)
        else df[key].eq(val).fillna(value=False)  # type: ignore
        for key, val in criteria.items()
    ]
    for (key, val), is_eligible_citerion in zip(criteria.items(), is_eligibles):
        nuq_students_criterion = (
            df.loc[is_eligible_citerion, student_id_cols]
            .groupby(by=student_id_cols, sort=False)
            .ngroups
        )
        _log_eligible_selection(
            nuq_students_in, nuq_students_criterion, f"{key}={val} criterion ..."
        )

    is_eligible = np.logical_and.reduce(is_eligibles)
    df_out = (
        df.loc[is_eligible, student_id_cols]
        # df is at student-term level; get student ids from first eligible terms only
        .drop_duplicates(ignore_index=True)
    )
    _log_eligible_selection(nuq_students_in, len(df_out), "all criteria")
    return df_out


def select_students_by_time_left(
    df: pd.DataFrame,
    *,
    intensity_time_lefts: list[tuple[str, float, t.Literal["year", "term"]]],
    max_term_rank: int,
    num_terms_in_year: int = 4,
    student_id_cols: str | list[str] = "student_guid",
    enrollment_intensity_col: str = "enrollment_intensity_first_term",
    term_rank_col: str = "term_rank",
) -> pd.DataFrame:
    """
    Select distinct students in ``df`` for which ``df`` includes enough time left
    (in terms or years) to cover the student's future "target" term,
    depending on the student's enrollment intensity.

    Args:
        df: Student-term dataset. See "Warning" below.
        intensity_time_lefts: One or more criteria that define the amount of time left
            for a given enrollment intensity for a student to be considered eligible,
            specified as a triple: (enrollment intensity, amount of time, time unit).
            For example:

                - ``[("FULL-TIME", 1.0, "year")]`` => for students enrolled at full-time
                  intensity, at least 1 year's worth of student-terms must be included
                  in the full dataset following the given student-term
                - ``[("FULL-TIME", 4.0, "term"), ("PART-TIME", 8.0, "term)]`` =>
                  for full-time students, at least 4 subsequent terms must be included
                  in the dataset; for part-time students, at least 8 subsequent terms

        max_term_rank: Maximum "term rank" value in full dataset. Not necessarily
            the maximum term rank included in ``df`` , depending on how it's filtered.
        num_terms_in_year: Number of (academic) terms in a(n academic) year.
            Used to convert times given in "year" units into times in "term" units.
        student_id_cols: Column(s) that uniquely identify students in ``df`` ,
            used to drop duplicates.
        enrollment_intensity_col: Column name in ``df`` with enrollment intensity values.
        term_rank_col: Column name in ``df`` with term rank values.

    Warning:
        This assumes that ``df`` only includes one student-term row per student; namely,
        the one from which time left should be measured. If that assumption is violated,
        this code actually checks if *any* of the student's terms occurred with
        enough time left for their particular enrollment intensity.
    """
    student_id_cols = utils.to_list(student_id_cols)
    nuq_students_in = df.groupby(by=student_id_cols, sort=False).ngroups
    intensity_num_terms = _compute_intensity_num_terms(
        intensity_time_lefts, num_terms_in_year
    )
    has_enough_terms_lefts = [
        (
            # enrollment intensity is equal to a specified value
            df[enrollment_intensity_col].eq(intensity)
            # and at least X terms left between reference term and max term in dataset
            & (max_term_rank - df[term_rank_col]).ge(num_terms)
        )
        for intensity, num_terms in intensity_num_terms
    ]
    has_enough_terms_left = np.logical_or.reduce(has_enough_terms_lefts)
    df_out = (
        df.loc[has_enough_terms_left, student_id_cols]
        # df is at student-term level; get student ids from first eligible terms only
        .drop_duplicates(ignore_index=True)
    )
    _log_eligible_selection(nuq_students_in, len(df_out), "time left")
    return df_out


def select_students_by_next_year_course_data(
    df: pd.DataFrame,
    *,
    student_id_cols: str | list[str] = "student_guid",
    cohort_id_col: str = "cohort_id",
    term_id_col: str = "term_id",
) -> pd.DataFrame:
    """
    Select distinct students in ``df`` for which ``df`` includes any records
    for the *next* academic year after the students' cohort year; effectively,
    this drops all students in the cohort from the most recent course year.

    Args:
        df: Student-term dataset.
        student_id_cols
        cohort_id_col
        term_id_col
    """
    student_id_cols = utils.to_list(student_id_cols)
    nuq_students_in = df.groupby(by=student_id_cols, sort=False).ngroups
    max_term_year = (
        df[term_id_col].str.extract(r"^(\d{4})").astype("Int32").max().iat[0]
    )
    df_out = (
        df.groupby(by=student_id_cols, as_index=False)
        .agg(student_cohort_id=(cohort_id_col, "min"))
        .assign(
            student_cohort_year=lambda df: df["student_cohort_id"]
            .str.extract(r"^(\d{4})")
            .astype("Int32")
        )
        .loc[lambda df: df["student_cohort_year"] < max_term_year, student_id_cols]
    )
    _log_eligible_selection(nuq_students_in, len(df_out), "next year course data")
    return df_out


def get_first_student_terms(
    df: pd.DataFrame,
    *,
    student_id_cols: str | list[str] = "student_guid",
    sort_cols: str | list[str] = "term_rank",
    include_cols: t.Optional[list[str]] = None,
) -> pd.DataFrame:
    """
    For each student, get the first row in ``df`` (in ascending order of ``sort_cols`` )
    and a configurable subset of columns.

    Args:
        df
        student_id_cols
        sort_cols
        include_cols
    """
    student_id_cols = utils.to_list(student_id_cols)
    sort_cols = utils.to_list(sort_cols)
    cols = (
        df.columns.tolist()
        if include_cols is None
        else list(
            utils.unique_elements_in_order(student_id_cols + sort_cols + include_cols)
        )
    )
    return (
        df.loc[:, cols]
        .sort_values(by=sort_cols, ascending=True)
        .groupby(by=student_id_cols, sort=False, as_index=False)
        .first()
    )


def get_first_student_terms_at_num_credits_earned(
    df: pd.DataFrame,
    *,
    min_num_credits: float,
    student_id_cols: str | list[str] = "student_guid",
    sort_cols: str | list[str] = "term_rank",
    num_credits_col: str = "num_credits_earned_cumsum",
    include_cols: t.Optional[list[str]] = None,
) -> pd.DataFrame:
    """
    For each student, get the first row in ``df`` (in ascending order of ``sort_cols`` )
    for which their cumulative num credits earned is greater than or equal
    to the specified threshold value, and drop students that never earn enough.

    Args:
        df
        min_num_credits
        student_id_cols
        sort_cols
        num_credits_col
        include_cols
    """
    return get_first_student_terms(
        # exclude rows with insufficient num credits, so "first" meets our criteria here
        df.loc[df[num_credits_col].ge(min_num_credits), :],
        student_id_cols=student_id_cols,
        sort_cols=sort_cols,
        include_cols=include_cols,
    )


def get_first_student_terms_within_cohort(
    df: pd.DataFrame,
    *,
    student_id_cols: str | list[str] = "student_guid",
    term_is_pre_cohort_col: str = "term_is_pre_cohort",
    sort_cols: str | list[str] = "term_rank",
    include_cols: t.Optional[list[str]] = None,
) -> pd.DataFrame:
    """
    For each student, get the first row in ``df`` (in ascending order of ``sort_cols`` )
    for which the term occurred *within* the student's cohort, i.e. not prior to
    their official start of enrollment.

    Args:
        df
        student_id_cols
        cohort_id_col
        term_id_col
        term_rank_col
        sort_cols
        include_cols
    """
    return get_first_student_terms(
        # exclude rows that are "pre-cohort", so "first" meets our criteria here
        df.loc[df[term_is_pre_cohort_col].eq(False), :],
        student_id_cols=student_id_cols,
        sort_cols=sort_cols,
        include_cols=include_cols,
    )


def _compute_intensity_num_terms(
    intensity_time_lefts: list[tuple[str, float, t.Literal["year", "term"]]],
    num_terms_in_year: int,
) -> list[tuple[str, float]]:
    """
    Compute the minimum number of terms required for a given enrollment intensity,
    converting values given in "years" into "terms" as needed.
    """
    return [
        (intensity, time if unit == "term" else time * num_terms_in_year)
        for intensity, time, unit in intensity_time_lefts
    ]


def _log_eligible_selection(
    nunique_students_in: int, nunique_students_out: int, case: str
) -> None:
    LOGGER.info(
        "%s out of %s (%s%%) students selected as eligible by %s",
        nunique_students_out,
        nunique_students_in,
        round(100 * nunique_students_out / nunique_students_in, 1),
        case,
    )



import pandas as pd
import pytest

from student_success_tool.analysis.pdp.targets import shared


@pytest.fixture(scope="module")
def test_df():
    return pd.DataFrame(
        {
            "student_id": ["01", "01", "01", "02", "02", "03", "04", "05"],
            "cohort_id": [
                "2020-21 SPRING",
                "2020-21 SPRING",
                "2020-21 SPRING",
                "2021-22 FALL",
                "2021-22 FALL",
                "2019-20 FALL",
                "2020-21 SPRING",
                "2022-23 FALL",
            ],
            "term_id": [
                "2020-21 FALL",
                "2020-21 SPRING",
                "2021-22 FALL",
                "2021-22 FALL",
                "2021-22 SPRING",
                "2019-20 SPRING",
                "2020-21 SPRING",
                "2022-23 FALL",
            ],
            "credential_sought": [
                "Associate's",
                "Associate's",
                "Associate's",
                "Bachelor's",
                "Bachelor's",
                "Associate's",
                "Associate's",
                pd.NA,
            ],
            "enrollment_type": [
                "FIRST-TIME",
                "FIRST-TIME",
                "FIRST-TIME",
                "FIRST-TIME",
                "FIRST-TIME",
                "FIRST-TIME",
                "TRANSFER-IN",
                pd.NA,
            ],
            "enrollment_intensity": [
                "FULL-TIME",
                "FULL-TIME",
                "FULL-TIME",
                "PART-TIME",
                "PART-TIME",
                "PART-TIME",
                "FULL-TIME",
                pd.NA,
            ],
            "num_credits_earned": [25.0, 30.0, 35.0, 25.0, 35.0, 20.0, 45.0, 10.0],
            "term_rank": [3, 4, 5, 5, 6, 2, 4, 8],
            "term_is_pre_cohort": [
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        },
    ).astype(
        {
            "student_id": "string",
            "credential_sought": "string",
            "enrollment_type": "string",
            "enrollment_intensity": "string",
        }
    )


@pytest.mark.parametrize(
    ["criteria", "exp"],
    [
        (
            {
                "credential_sought": "Associate's",
                "enrollment_type": "FIRST-TIME",
                "enrollment_intensity": {"FULL-TIME", "PART-TIME"},
            },
            pd.DataFrame({"student_id": ["01", "03"]}),
        ),
        (
            {
                "enrollment_type": "FIRST-TIME",
                "enrollment_intensity": "PART-TIME",
            },
            pd.DataFrame({"student_id": ["02", "03"]}),
        ),
        (
            {"credential_sought": ["Associate's", "Bachelor's"]},
            pd.DataFrame({"student_id": ["01", "02", "03", "04"]}),
        ),
        (
            {"enrollment_type": "RE-ADMIT"},
            pd.DataFrame({"student_id": []}),
        ),
    ],
)
def test_select_students_by_criteria(test_df, criteria, exp):
    obs = shared.select_students_by_criteria(
        test_df, student_id_cols="student_id", **criteria
    )
    assert isinstance(obs, pd.DataFrame)
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["intensity_time_lefts", "max_term_rank", "num_terms_in_year", "exp"],
    [
        (
            [("FULL-TIME", 1.0, "year"), ("PART-TIME", 2.0, "year")],
            8,
            4,
            pd.DataFrame({"student_id": ["01", "04"]}),
        ),
        (
            [("FULL-TIME", 4.0, "term"), ("PART-TIME", 8.0, "term")],
            8,
            4,
            pd.DataFrame({"student_id": ["01", "04"]}),
        ),
        (
            [("FULL-TIME", 1.0, "year"), ("PART-TIME", 2.0, "year")],
            10,
            4,
            pd.DataFrame({"student_id": ["01", "03", "04"]}),
        ),
        (
            [("FULL-TIME", 1.0, "year")],
            8,
            4,
            pd.DataFrame({"student_id": ["01", "04"]}),
        ),
        (
            [("PART-TIME", 2.0, "year")],
            8,
            4,
            pd.DataFrame({"student_id": []}),
        ),
        (
            [("FULL-TIME", 1.0, "year"), ("PART-TIME", 2.0, "year")],
            8,
            3,
            pd.DataFrame({"student_id": ["01", "03", "04"]}),
        ),
    ],
)
def test_select_students_by_time_left(
    test_df, intensity_time_lefts, max_term_rank, num_terms_in_year, exp
):
    obs = shared.select_students_by_time_left(
        test_df,
        intensity_time_lefts=intensity_time_lefts,
        max_term_rank=max_term_rank,
        num_terms_in_year=num_terms_in_year,
        student_id_cols="student_id",
        enrollment_intensity_col="enrollment_intensity",
        term_rank_col="term_rank",
    )
    assert isinstance(obs, pd.DataFrame)
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    "exp",
    [
        pd.DataFrame({"student_id": ["01", "02", "03", "04"]}),
    ],
)
def test_select_students_by_next_year_course_data(test_df, exp):
    obs = shared.select_students_by_next_year_course_data(
        test_df, student_id_cols="student_id"
    )
    assert isinstance(obs, pd.DataFrame)
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["include_cols", "exp"],
    [
        (
            [],
            pd.DataFrame(
                {
                    "student_id": ["03", "01", "04", "02", "05"],
                    "term_rank": [2, 3, 4, 5, 8],
                }
            ),
        ),
        (
            ["num_credits_earned"],
            pd.DataFrame(
                {
                    "student_id": ["03", "01", "04", "02", "05"],
                    "term_rank": [2, 3, 4, 5, 8],
                    "num_credits_earned": [20.0, 25.0, 45.0, 25.0, 10.0],
                }
            ),
        ),
        (
            ["num_credits_earned", "term_rank", "student_id"],
            pd.DataFrame(
                {
                    "student_id": ["03", "01", "04", "02", "05"],
                    "term_rank": [2, 3, 4, 5, 8],
                    "num_credits_earned": [20.0, 25.0, 45.0, 25.0, 10.0],
                }
            ),
        ),
        (
            None,
            pd.DataFrame(
                {
                    "student_id": ["03", "01", "04", "02", "05"],
                    "cohort_id": [
                        "2019-20 FALL",
                        "2020-21 SPRING",
                        "2020-21 SPRING",
                        "2021-22 FALL",
                        "2022-23 FALL",
                    ],
                    "term_id": [
                        "2019-20 SPRING",
                        "2020-21 FALL",
                        "2020-21 SPRING",
                        "2021-22 FALL",
                        "2022-23 FALL",
                    ],
                    "credential_sought": [
                        "Associate's",
                        "Associate's",
                        "Associate's",
                        "Bachelor's",
                        pd.NA,
                    ],
                    "enrollment_type": [
                        "FIRST-TIME",
                        "FIRST-TIME",
                        "TRANSFER-IN",
                        "FIRST-TIME",
                        pd.NA,
                    ],
                    "enrollment_intensity": [
                        "PART-TIME",
                        "FULL-TIME",
                        "FULL-TIME",
                        "PART-TIME",
                        pd.NA,
                    ],
                    "num_credits_earned": [20.0, 25.0, 45.0, 25.0, 10.0],
                    "term_rank": [2, 3, 4, 5, 8],
                    "term_is_pre_cohort": [False, True, False, False, False],
                }
            ).astype(
                {
                    "credential_sought": "string",
                    "enrollment_type": "string",
                    "enrollment_intensity": "string",
                }
            ),
        ),
    ],
)
def test_get_first_student_terms(test_df, include_cols, exp):
    obs = shared.get_first_student_terms(
        test_df,
        student_id_cols="student_id",
        sort_cols="term_rank",
        include_cols=include_cols,
    )
    assert isinstance(obs, pd.DataFrame)
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["min_num_credits", "include_cols", "exp"],
    [
        (
            10.0,
            [],
            pd.DataFrame(
                {
                    "student_id": ["03", "01", "04", "02", "05"],
                    "term_rank": [2, 3, 4, 5, 8],
                }
            ),
        ),
        (
            30.0,
            ["num_credits_earned"],
            pd.DataFrame(
                {
                    "student_id": ["01", "04", "02"],
                    "term_rank": [4, 4, 6],
                    "num_credits_earned": [30.0, 45.0, 35.0],
                }
            ),
        ),
        (
            60.0,
            [],
            pd.DataFrame({"student_id": [], "term_rank": []}),
        ),
    ],
)
def test_get_first_student_terms_at_num_credits_earned(
    test_df, min_num_credits, include_cols, exp
):
    obs = shared.get_first_student_terms_at_num_credits_earned(
        test_df,
        min_num_credits=min_num_credits,
        student_id_cols="student_id",
        sort_cols="term_rank",
        num_credits_col="num_credits_earned",
        include_cols=include_cols,
    )
    assert isinstance(obs, pd.DataFrame)
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["include_cols", "exp"],
    [
        (
            [],
            pd.DataFrame(
                {
                    "student_id": ["03", "01", "04", "02", "05"],
                    "term_rank": [2, 4, 4, 5, 8],
                }
            ),
        ),
        (
            ["cohort_id", "term_id"],
            pd.DataFrame(
                {
                    "student_id": ["03", "01", "04", "02", "05"],
                    "term_rank": [2, 4, 4, 5, 8],
                    "cohort_id": [
                        "2019-20 FALL",
                        "2020-21 SPRING",
                        "2020-21 SPRING",
                        "2021-22 FALL",
                        "2022-23 FALL",
                    ],
                    "term_id": [
                        "2019-20 SPRING",
                        "2020-21 SPRING",
                        "2020-21 SPRING",
                        "2021-22 FALL",
                        "2022-23 FALL",
                    ],
                }
            ),
        ),
    ],
)
def test_get_first_student_terms_within_cohort(test_df, include_cols, exp):
    obs = shared.get_first_student_terms_within_cohort(
        test_df,
        student_id_cols="student_id",
        sort_cols="term_rank",
        include_cols=include_cols,
    )
    assert isinstance(obs, pd.DataFrame)
    assert obs.equals(exp) or obs.compare(exp).empty



import re
import typing as t
from collections.abc import Collection, Iterable

RE_VARIOUS_PUNCTS = re.compile(r"[!()*+\,\-./:;<=>?[\]^_{|}~]")
RE_QUOTATION_MARKS = re.compile(r"[\'\"\`]")


def to_list(value: t.Any) -> list:
    """Cast ``value`` into a list, regardless of its type."""
    if isinstance(value, list):
        return value
    elif isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
        return list(value)
    else:
        return [value]


def is_collection_but_not_string(value: t.Any) -> bool:
    """Check if ``value`` is a collection, excluding strings (and bytes)."""
    return isinstance(value, Collection) and not isinstance(value, (str, bytes))


def unique_elements_in_order(eles: Iterable) -> Iterable:
    """Get unique elements from an iterable, in order of appearance."""
    seen = set()  # type: ignore
    seen_add = seen.add
    for ele in eles:
        if ele not in seen:
            seen_add(ele)
            yield ele




def convert_to_snake_case(col: str) -> str:
    """Convert column name into snake case, without punctuation."""
    col = RE_VARIOUS_PUNCTS.sub(" ", col)
    col = RE_QUOTATION_MARKS.sub("", col)
    # TODO: *pretty sure* this could be cleaner and more performant, but shrug
    words = re.sub(
        r"([A-Z][a-z]+)", r" \1", re.sub(r"([A-Z]+|[0-9]+|\W+)", r" \1", col)
    ).split()
    return "_".join(words).lower()



import logging
from collections.abc import Collection

import pandas as pd

from .. import utils
from . import shared

LOGGER = logging.getLogger(__name__)


def make_labeled_dataset(
    df: pd.DataFrame,
    *,
    student_criteria: dict[str, object | Collection[object]],
    exclude_pre_cohort_terms: bool = True,
    student_id_cols: str | list[str] = "student_guid",
    cohort_id_col: str = "cohort_id",
    term_id_col: str = "term_id",
    term_is_pre_cohort_col: str = "term_is_pre_cohort",
    term_rank_col: str = "term_rank",
    retention_col: str = "retention",
) -> pd.DataFrame:
    """
    Make a labeled dataset for modeling, where each row consists of features
    from eligible students' first term (within their cohort)
    matched to computed target variables.

    See Also:
        - :func:`select_eligible_students()`
        - :func:`compute_target_variable()`
        - :func:`shared.get_first_student_terms_within_cohort()`
    """
    df_eligible_students = select_eligible_students(
        df,
        student_criteria=student_criteria,
        student_id_cols=student_id_cols,
        cohort_id_col=cohort_id_col,
        term_id_col=term_id_col,
    )
    df_eligible_student_terms = pd.merge(
        df, df_eligible_students, on=student_id_cols, how="inner"
    )
    if exclude_pre_cohort_terms is True:
        df_features = shared.get_first_student_terms_within_cohort(
            df_eligible_student_terms,
            student_id_cols=student_id_cols,
            term_is_pre_cohort_col=term_is_pre_cohort_col,
            sort_cols=term_rank_col,
            include_cols=None,
        )
    else:
        df_features = shared.get_first_student_terms(
            df_eligible_student_terms,
            student_id_cols=student_id_cols,
            sort_cols=term_rank_col,
            include_cols=None,
        )
    df_targets = compute_target_variable(
        df_eligible_student_terms,
        student_id_cols=student_id_cols,
        retention_col=retention_col,
    )
    df_labeled = pd.merge(df_features, df_targets, on=student_id_cols, how="inner")
    return df_labeled


def compute_target_variable(
    df: pd.DataFrame,
    *,
    student_id_cols: str | list[str] = "student_guid",
    retention_col: str = "retention",
) -> pd.Series:
    """
    Args:
        df
        student_id_cols
        retention_col
    """
    student_id_cols = utils.to_list(student_id_cols)
    return (
        df[student_id_cols + [retention_col]]
        .drop_duplicates(subset=student_id_cols, ignore_index=True)
        # recall that we are prediction *non* retention
        .assign(target=lambda df: ~df[retention_col])
        .drop(columns=retention_col)
        .fillna({"target": False})
        .astype({"target": "bool"})
        # TODO: do we want a series with student ids as index and target as values, or nah?
        .set_index(student_id_cols)
        .loc[:, "target"]
    )


def select_eligible_students(
    df: pd.DataFrame,
    *,
    student_criteria: dict[str, object | Collection[object]],
    student_id_cols: str | list[str] = "student_guid",
    cohort_id_col: str = "cohort_id",
    term_id_col: str = "term_id",
) -> pd.DataFrame:
    """
    Args:
        df
        student_criteria
        student_id_cols
        cohort_id_col
        term_id_col
    """
    nuq_students_in = df.groupby(by=student_id_cols, sort=False).ngroups
    df_students_by_criteria = shared.select_students_by_criteria(
        df, student_id_cols=student_id_cols, **student_criteria
    )
    df_students_by_next_year_course_data = (
        shared.select_students_by_next_year_course_data(
            df,
            student_id_cols=student_id_cols,
            cohort_id_col=cohort_id_col,
            term_id_col=term_id_col,
        )
    )
    df_out = pd.merge(
        df_students_by_criteria,
        df_students_by_next_year_course_data,
        on=student_id_cols,
        how="inner",
    )
    nuq_students_out = df_out.groupby(by=student_id_cols, sort=False).ngroups
    shared._log_eligible_selection(nuq_students_in, nuq_students_out, "overall")
    return df_out



import functools as ft
import logging
import time
import typing as t

import pandas as pd
import pandera as pda
import pandera.errors
import pyspark.sql

from . import utils

LOGGER = logging.getLogger(__name__)


def read_raw_pdp_course_data_from_file(
    fpath: str,
    *,
    schema: t.Optional[pda.DataFrameModel],
    dttm_format: str = "%Y%m%d.0",
    preprocess_func: t.Optional[t.Callable[[pd.DataFrame], pd.DataFrame]] = None,
    **kwargs: object,
) -> pd.DataFrame:
    """
    Read a raw PDP course data file stored on disk at ``fpath`` in CSV format,
    and parse+validate it via ``schema`` .

    Args:
        file_path
        schema: "DataFrameModel", such as those specified in :mod:`pdp.schemas` ,
            used to parse and validate the raw data. If None, parsing/validation
            is skipped, and the raw data is returned as-is.
        dttm_format: Datetime format for "Course Begin/End Date" columns.
        preprocess_func: If the raw data at ``fpath`` is incompatible with ``schema`` ,
            provide a function that takes the raw dataframe as its sole input,
            performs whatever (minimal) transformations necessary to bring the data
            into line with ``schema`` , and then returns it. This preprocessed dataset
            will then be passed into ``schema`` , if specified.
            NOTE: Allowances for minor differences in the data should be implemented
            on the school-specific schema class directly. This function is intended
            to handle bigger problems, such as dupe ids or borked columns.
        **kwargs: Additional arguments passed as-is into :func:`pd.read_csv()` .
            Note that ``dtype`` is hard-coded, so you can't specify it again here.

    References:
        - https://help.studentclearinghouse.org/pdp/knowledge-base/course-level-analysis-ready-file-data-dictionary
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        - https://pandera.readthedocs.io/en/stable/reference/generated/pandera.api.dataframe.model.DataFrameModel.html#pandera.api.dataframe.model.DataFrameModel.validate
    """
    df = (
        # read everything as a string to start, for maximum flexibility
        pd.read_csv(fpath, dtype="string", **kwargs)  # type: ignore
        # make minimal changes before parsing/validating via schema
        # standardize column names, for convenience/consistency
        .rename(columns=utils.convert_to_snake_case)
        # basically, any operations needed for dtype coercion to work correctly
        .assign(
            # uppercase string values for some cols to avoid case inconsistency later on
            **{
                col: ft.partial(_uppercase_string_values, col=col)
                for col in ("academic_term", "student_age", "race")
            }
            # help pandas to parse non-standard datetimes... read_csv() struggles
            | {
                col: ft.partial(_parse_dttm_values, col=col, fmt=dttm_format)
                for col in ("course_begin_date", "course_end_date")
            }
        )
    )
    LOGGER.info("read course data from %s", fpath)
    assert isinstance(df, pd.DataFrame)  # type guard
    if preprocess_func is not None:
        LOGGER.info("applying %s preprocessor to raw dataset", preprocess_func)
        df = preprocess_func(df)
    if schema is None:
        return df
    else:
        try:
            df = schema.validate(df, lazy=True)
            return df  # type: ignore
        except pandera.errors.SchemaErrors:
            LOGGER.error("unable to parse/validate course data")
            raise


def read_raw_pdp_cohort_data_from_file(
    fpath: str,
    *,
    schema: t.Optional[pda.DataFrameModel],
    preprocess_func: t.Optional[t.Callable[[pd.DataFrame], pd.DataFrame]] = None,
    **kwargs: object,
) -> pd.DataFrame:
    """
    Read a raw PDP cohort data file stored on disk at ``fpath`` in CSV format.

    Args:
        fpath
        schema: "DataFrameModel", such as those specified in :mod:`pdp.schemas` ,
            used to parse and validate the raw data. If None, parsing/validation
            is skipped, and the raw data is returned as-is.
        preprocess_func: If the raw data at ``fpath`` is incompatible with ``schema`` ,
            provide a function that takes the raw dataframe as its sole input,
            performs whatever (minimal) transformations necessary to bring the data
            into line with ``schema`` , and then returns it. This preprocessed dataset
            will then be passed into ``schema`` , if specified.
            NOTE: Allowances for minor differences in the data should be implemented
            on the school-specific schema class directly. This function is intended
            to handle bigger problems, such as dupe ids or borked columns.
        **kwargs: Additional arguments passed as-is into :func:`pd.read_csv()` .
            Note that ``dtype`` is hard-coded to "string", so you can't specify it here.

    References:
        - https://help.studentclearinghouse.org/pdp/knowledge-base/cohort-level-analysis-ready-file-data-dictionary
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        - https://pandera.readthedocs.io/en/stable/reference/generated/pandera.api.dataframe.model.DataFrameModel.html#pandera.api.dataframe.model.DataFrameModel.validate
    """
    df = (
        # read everything as a string to start, for maximum flexibility
        pd.read_csv(fpath, dtype="string", **kwargs)  # type: ignore
        # make minimal changes before parsing/validating via schema
        # standardize column names, for convenience/consistency
        .rename(columns=utils.convert_to_snake_case)
        # basically, any operations needed for dtype coercion to work correctly
        .assign(
            # uppercase string values for some cols to avoid case inconsistency later on
            # for practical reasons, this is the only place where it's easy to do so
            **{
                col: ft.partial(_uppercase_string_values, col=col)
                for col in (
                    "cohort_term",
                    "enrollment_type",
                    "enrollment_intensity_first_term",
                    "student_age",
                    "race",
                    "most_recent_bachelors_at_other_institution_locale",
                    "most_recent_associates_or_certificate_at_other_institution_locale",
                    "most_recent_last_enrollment_at_other_institution_locale",
                    "first_bachelors_at_other_institution_locale",
                    "first_associates_or_certificate_at_other_institution_locale",
                )
            }
            # replace "UK" with null in GPA cols, so we can coerce to float via schema
            | {
                col: ft.partial(_replace_values_with_null, col=col, to_replace="UK")
                for col in ("gpa_group_term_1", "gpa_group_year_1")
            }
            # help pandas to coerce string "1"/"0" values into True/False
            | {
                col: ft.partial(_cast_to_bool_via_int, col=col)
                for col in ("retention", "persistence")
            }
        )
    )
    LOGGER.info("read cohort data from %s", fpath)
    assert isinstance(df, pd.DataFrame)  # type guard
    if preprocess_func is not None:
        LOGGER.info("applying %s preprocessor to raw dataset", preprocess_func)
        df = preprocess_func(df)
    if schema is None:
        return df
    else:
        try:
            df = schema.validate(df, lazy=True)
            return df  # type: ignore
        except pandera.errors.SchemaErrors:
            LOGGER.error("unable to parse/validate cohort data")
            raise


def _parse_dttm_values(df: pd.DataFrame, *, col: str, fmt: str) -> pd.Series:
    return pd.to_datetime(df[col], format=fmt)


def _uppercase_string_values(df: pd.DataFrame, *, col: str) -> pd.Series:
    return df[col].str.upper()


def _replace_values_with_null(
    df: pd.DataFrame, *, col: str, to_replace: str | list[str]
) -> pd.Series:
    return df[col].replace(to_replace=to_replace, value=None)


def _cast_to_bool_via_int(df: pd.DataFrame, *, col: str) -> pd.Series:
    return df[col].astype("Int8").astype("boolean")


def read_data_from_delta_table(
    table_path: str, spark_session: pyspark.sql.SparkSession
) -> pd.DataFrame:
    """
    Read data from a table in Databricks Unity Catalog and return it as a DataFrame.

    Args:
        table_path: Path in Unity Catalog from which data will be read,
            including the full three-level namespace: ``catalog.schema.table`` .
        spark_session: Entry point to using spark dataframes and the databricks integration.

    See Also:
        - https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.html#pyspark.sql.DataFrameReader
        - https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/spark_session.html
    """
    df = spark_session.read.format("delta").table(table_path).toPandas()
    assert isinstance(df, pd.DataFrame)  # type guard
    return df


def write_data_to_delta_table(
    df: pd.DataFrame, table_path: str, spark_session: pyspark.sql.SparkSession
) -> None:
    """
    Write pandas DataFrame to Databricks Unity Catalog.

    Args:
        df
        table_path: Path in Unity Catalog to which ``df`` will be written,
            including the full three-level namespace: ``catalog.schema.table`` .
        spark_session: Entry point to using spark dataframes and the databricks integration.

    See Also:
        - https://docs.databricks.com/en/delta/drop-table.html#when-to-replace-a-table
        - https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/spark_session.html
    """
    start_time = time.time()
    LOGGER.info("saving data to %s ...", table_path)
    df_spark = spark_session.createDataFrame(
        df.rename(columns=utils.convert_to_snake_case)
    )
    (
        pyspark.sql.DataFrameWriterV2(df_spark, table_path)
        .options(format="delta")
        # this *should* do what databricks recomends -- and retains table history!
        .createOrReplace()
    )
    run_time = time.time() - start_time

    table_rows = spark_session.sql(f"SELECT COUNT(*) FROM {table_path}").collect()
    if table_rows[0][0] != len(df):
        raise IOError(
            f"{table_rows[0][0]} written to delta table, "
            f"but {len(df)} rows in original dataframe"
        )

    history = spark_session.sql(f"DESCRIBE history {table_path} LIMIT 1").collect()
    verno = int(history[0][0])
    LOGGER.info("data saved to %s (v%s) in %s seconds", table_path, verno, run_time)
